# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import datetime
from copy import deepcopy
from typing import List, Union

import numpy as np
import pytest

from otx.api.configuration import ConfigurableParameters
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.color import Color
from otx.api.entities.dataset_item import DatasetItemEntity, DatasetItemEntityWithID
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.id import ID
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.metadata import MetadataItemEntity
from otx.api.entities.model import ModelConfiguration, ModelEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.ellipse import Ellipse
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.subset import Subset
from otx.api.entities.tensor import TensorEntity
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


class DatasetItemParameters:
    @staticmethod
    def generate_random_image() -> Image:
        image = Image(data=np.random.randint(low=0, high=255, size=(10, 16, 3)))
        return image

    @staticmethod
    def labels() -> List[LabelEntity]:
        creation_date = datetime.datetime(year=2021, month=12, day=9)
        detection_label = LabelEntity(
            name="Label for Detection",
            domain=Domain.DETECTION,
            color=Color(red=100, green=200, blue=150),
            creation_date=creation_date,
            id=ID("detection_label"),
        )
        segmentation_label = LabelEntity(
            name="Label for Segmentation",
            domain=Domain.DETECTION,
            color=Color(red=50, green=80, blue=200),
            creation_date=creation_date,
            is_empty=True,
            id=ID("segmentation_label"),
        )
        return [detection_label, segmentation_label]

    def annotations(self) -> List[Annotation]:
        labels = self.labels()
        rectangle = Rectangle(x1=0.2, y1=0.2, x2=0.6, y2=0.7)
        other_rectangle = Rectangle(x1=0.3, y1=0.2, x2=0.9, y2=0.9)
        detection_annotation = Annotation(
            shape=rectangle,
            labels=[ScoredLabel(label=labels[0])],
            id=ID("detection_annotation_1"),
        )
        segmentation_annotation = Annotation(
            shape=other_rectangle,
            labels=[ScoredLabel(label=labels[1])],
            id=ID("segmentation_annotation_1"),
        )
        return [detection_annotation, segmentation_annotation]

    @staticmethod
    def roi_labels() -> List[LabelEntity]:
        creation_date = datetime.datetime(year=2021, month=12, day=9)
        roi_label = LabelEntity(
            name="ROI label",
            domain=Domain.DETECTION,
            color=Color(red=40, green=180, blue=80),
            creation_date=creation_date,
            id=ID("roi_label_1"),
        )
        other_roi_label = LabelEntity(
            name="Second ROI label",
            domain=Domain.SEGMENTATION,
            color=Color(red=80, green=90, blue=70),
            creation_date=creation_date,
            is_empty=True,
            id=ID("roi_label_2"),
        )
        return [roi_label, other_roi_label]

    def roi_scored_labels(self) -> List[ScoredLabel]:
        roi_labels = self.roi_labels()
        return [ScoredLabel(roi_labels[0]), ScoredLabel(roi_labels[1])]

    def roi(self):
        roi = Annotation(
            shape=Rectangle(
                x1=0.1,
                y1=0.1,
                x2=0.9,
                y2=0.9,
                modification_date=datetime.datetime(year=2021, month=12, day=9),
            ),
            labels=self.roi_scored_labels(),
            id=ID("roi_annotation"),
        )
        return roi

    @staticmethod
    def metadata() -> List[MetadataItemEntity]:
        data = TensorEntity(
            name="test_metadata",
            numpy=np.random.uniform(low=0.0, high=255.0, size=(10, 15, 3)),
        )
        other_data = TensorEntity(
            name="other_metadata",
            numpy=np.random.uniform(low=0.0, high=255.0, size=(10, 15, 3)),
        )
        return [MetadataItemEntity(data=data), MetadataItemEntity(data=other_data)]

    def annotations_entity(self) -> AnnotationSceneEntity:
        return AnnotationSceneEntity(
            annotations=self.annotations(),
            kind=AnnotationSceneKind.ANNOTATION,
            creation_date=datetime.datetime(year=2021, month=12, day=19),
            id=ID("annotation_entity_1"),
        )

    def default_values_dataset_item(self) -> DatasetItemEntity:
        return DatasetItemEntity(self.generate_random_image(), self.annotations_entity())

    def dataset_item(self) -> DatasetItemEntity:
        return DatasetItemEntity(
            media=self.generate_random_image(),
            annotation_scene=self.annotations_entity(),
            roi=self.roi(),
            metadata=self.metadata(),
            subset=Subset.TESTING,
            ignored_labels={self.labels()[1]},
        )

    def dataset_item_with_id(self) -> DatasetItemEntityWithID:
        return DatasetItemEntityWithID(
            id_=ID("test"),
            media=self.generate_random_image(),
            annotation_scene=self.annotations_entity(),
            roi=self.roi(),
            metadata=self.metadata(),
            subset=Subset.TESTING,
            ignored_labels={self.labels()[1]},
        )


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestDatasetItemEntity:
    @staticmethod
    def compare_denormalized_annotations(actual_annotations, expected_annotations) -> None:
        assert len(actual_annotations) == len(expected_annotations)
        for index in range(len(expected_annotations)):
            actual_annotation = actual_annotations[index]
            expected_annotation = expected_annotations[index]
            # Redefining id and modification_date required because of new Annotation objects created after shape
            # denormalize
            actual_annotation.id_ = expected_annotation.id_
            actual_annotation.shape.modification_date = expected_annotation.shape.modification_date
            assert actual_annotation == expected_annotation

    @staticmethod
    def labels_to_add() -> List[LabelEntity]:
        label_to_add = LabelEntity(
            name="Label which will be added",
            domain=Domain.DETECTION,
            color=Color(red=60, green=120, blue=70),
            creation_date=datetime.datetime(year=2021, month=12, day=12),
            id=ID("label_to_add_1"),
        )
        other_label_to_add = LabelEntity(
            name="Other label to add",
            domain=Domain.SEGMENTATION,
            color=Color(red=80, green=70, blue=100),
            creation_date=datetime.datetime(year=2021, month=12, day=11),
            is_empty=True,
            id=ID("label_to_add_2"),
        )
        return [label_to_add, other_label_to_add]

    def annotations_to_add(self) -> List[Annotation]:
        labels_to_add = self.labels_to_add()
        annotation_to_add = Annotation(
            shape=Rectangle(x1=0.1, y1=0.1, x2=0.7, y2=0.8),
            labels=[ScoredLabel(label=labels_to_add[0])],
            id=ID("added_annotation_1"),
        )
        other_annotation_to_add = Annotation(
            shape=Rectangle(x1=0.2, y1=0.3, x2=0.8, y2=0.9),
            labels=[ScoredLabel(label=labels_to_add[1])],
            id=ID("added_annotation_2"),
        )
        return [annotation_to_add, other_annotation_to_add]

    @staticmethod
    def metadata_item_with_model() -> MetadataItemEntity:
        data = TensorEntity(
            name="appended_metadata_with_model",
            numpy=np.random.randint(low=0, high=255, size=(10, 15, 3)),
        )
        configuration = ModelConfiguration(
            configurable_parameters=ConfigurableParameters(header="Test Header"),
            label_schema=LabelSchemaEntity(),
        )
        model = ModelEntity(configuration=configuration, train_dataset=DatasetEntity())
        metadata_item_with_model = MetadataItemEntity(data=data, model=model)
        return metadata_item_with_model

    @staticmethod
    def check_roi_equal_annotation(dataset_item: DatasetItemEntity, expected_labels: list, include_empty=False) -> None:
        roi_annotation_in_scene = None
        for annotation in dataset_item.annotation_scene.annotations:
            if annotation == dataset_item.roi:
                assert annotation.get_labels(include_empty=include_empty) == expected_labels
                roi_annotation_in_scene = annotation
                break
        assert roi_annotation_in_scene

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_item_initialization(self):
        """
        <b>Description:</b>
        Check DatasetItemEntity class object initialization

        <b>Input data:</b>
        DatasetItemEntity class objects with specified "media", "annotation_scene", "roi", "metadata" and "subset"
        parameters

        <b>Expected results:</b>
        Test passes if attributes of DatasetItemEntity class object are equal to expected

        <b>Steps</b>
        1. Check attributes of DatasetItemEntity object initialized with default optional parameters
        2. Check attributes of DatasetItemEntity object initialized with specified optional parameters
        """
        media = DatasetItemParameters.generate_random_image()
        annotations_scene = DatasetItemParameters().annotations_entity()
        # Checking attributes of DatasetItemEntity object initialized with default optional parameters
        default_values_dataset_item = DatasetItemEntity(media, annotations_scene)
        assert default_values_dataset_item.media == media
        assert default_values_dataset_item.annotation_scene == annotations_scene
        assert not default_values_dataset_item.get_metadata()
        assert default_values_dataset_item.subset == Subset.NONE
        assert default_values_dataset_item.ignored_labels == set()
        # Checking attributes of DatasetItemEntity object initialized with specified optional parameters
        roi = DatasetItemParameters().roi()
        metadata = DatasetItemParameters.metadata
        subset = Subset.TESTING
        ignored_labels = set(DatasetItemParameters().labels())
        specified_values_dataset_item = DatasetItemEntity(
            media, annotations_scene, roi, metadata, subset, ignored_labels
        )
        assert specified_values_dataset_item.media == media
        assert specified_values_dataset_item.annotation_scene == annotations_scene
        assert specified_values_dataset_item.roi == roi
        assert specified_values_dataset_item.get_metadata() == metadata
        assert specified_values_dataset_item.subset == subset
        assert specified_values_dataset_item.ignored_labels == ignored_labels

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_item_repr(self):
        """
        <b>Description:</b>
        Check DatasetItemEntity class __repr__ method

        <b>Input data:</b>
        DatasetItemEntity class objects with specified "media", "annotation_scene", "roi", "metadata" and "subset"
        parameters

        <b>Expected results:</b>
        Test passes if value returned by __repr__ method is equal to expected

        <b>Steps</b>
        1. Check value returned by __repr__ method for  DatasetItemEntity object with default optional parameters
        2. Check value returned by __repr__ method for  DatasetItemEntity object with specified optional parameters
        """
        media = DatasetItemParameters.generate_random_image()
        annotation_scene = DatasetItemParameters().annotations_entity()
        # Checking __repr__ method for DatasetItemEntity object initialized with default optional parameters
        default_values_dataset_item = DatasetItemEntity(media, annotation_scene)
        generated_roi = default_values_dataset_item.roi

        assert repr(default_values_dataset_item) == (
            f"DatasetItemEntity(media=Image(with data, width=16, height=10), "
            f"annotation_scene={annotation_scene}, roi={generated_roi}, "
            f"subset=NONE), meta=[]"
        )
        # Checking __repr__ method for DatasetItemEntity object initialized with specified optional parameters
        roi = DatasetItemParameters().roi()
        metadata = DatasetItemParameters.metadata()
        subset = Subset.TESTING
        specified_values_dataset_item = DatasetItemEntity(media, annotation_scene, roi, metadata, subset)
        assert repr(specified_values_dataset_item) == (
            f"DatasetItemEntity(media=Image(with data, width=16, height=10), annotation_scene={annotation_scene}, "
            f"roi={roi}, subset=TESTING), meta={metadata}"
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_item_roi(self):
        """
        <b>Description:</b>
        Check DatasetItemEntity class "roi" property

        <b>Input data:</b>
        DatasetItemEntity class object with specified "media", "annotation_scene", "roi", "metadata" and "subset"
        parameters

        <b>Expected results:</b>
        Test passes if value returned by "roi" property is equal to expected

        <b>Steps</b>
        1. Check value returned by "roi" property for DatasetItemEntity with specified "roi" parameter
        2. Check value returned by "roi" property for DatasetItemEntity with not specified "roi" parameter
        3. Check value returned by "roi" property for DatasetItemEntity with not specified "roi" parameter but one
        of annotation objects in annotation_scene is equal to full Rectangle
        """
        media = DatasetItemParameters.generate_random_image()
        annotations = DatasetItemParameters().annotations()
        annotation_scene = DatasetItemParameters().annotations_entity()
        roi = DatasetItemParameters().roi()
        metadata = DatasetItemParameters.metadata()
        # Checking "roi" property for DatasetItemEntity with specified "roi" parameter
        specified_roi_dataset_item = DatasetItemParameters().dataset_item()
        assert specified_roi_dataset_item.roi == roi
        # Checking that "roi" property is equal to full_box for DatasetItemEntity with not specified "roi" parameter
        non_specified_roi_dataset_item = DatasetItemEntity(media, annotation_scene, metadata=metadata)
        default_roi = non_specified_roi_dataset_item.roi.shape
        assert isinstance(default_roi, Rectangle)
        assert Rectangle.is_full_box(default_roi)
        # Checking that "roi" property will be equal to full_box for DatasetItemEntity with not specified "roi" but one
        # of Annotation objects in annotation_scene is equal to full Rectangle
        full_box_label = LabelEntity("Full-box label", Domain.DETECTION, id=ID("full_box_label"))
        full_box_annotation = Annotation(Rectangle.generate_full_box(), [ScoredLabel(full_box_label)])
        annotations.append(full_box_annotation)
        annotation_scene.annotations.append(full_box_annotation)
        full_box_label_dataset_item = DatasetItemEntity(media, annotation_scene, metadata=metadata)
        assert full_box_label_dataset_item.roi is full_box_annotation

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_item_roi_numpy(self):
        """
        <b>Description:</b>
        Check DatasetItemEntity class "roi_numpy" method

        <b>Input data:</b>
        DatasetItemEntity class object with specified "media", "annotation_scene", "roi", "metadata" and "subset"
        parameters

        <b>Expected results:</b>
        Test passes if array returned by "roi_numpy" method is equal to expected

        <b>Steps</b>
        1. Check array returned by roi_numpy method with not specified "roi" parameter for DatasetItemEntity with
        "roi" attribute is "None"
        2. Check array returned by roi_numpy method with Rectangle-shape "roi" parameter
        3. Check array returned by roi_numpy method with Ellipse-shape "roi" parameter
        4. Check array returned by roi_numpy method with Polygon-shape "roi" parameter
        5. Check array returned by roi_numpy method with non-specified "roi" parameter for DatasetItemEntity with "roi"
        attribute
        """
        media = DatasetItemParameters.generate_random_image()
        annotation_scene = DatasetItemParameters().annotations_entity()
        roi_label = LabelEntity("ROI label", Domain.DETECTION, id=ID("roi_label"))
        dataset_item = DatasetItemEntity(media, annotation_scene)
        # Checking array returned by "roi_numpy" method with non-specified "roi" parameter for DatasetItemEntity
        # "roi" attribute is "None"
        assert np.array_equal(dataset_item.roi_numpy(), media.numpy)
        # Checking array returned by "roi_numpy" method with specified Rectangle-shape "roi" parameter
        rectangle_roi = Annotation(
            Rectangle(x1=0.2, y1=0.1, x2=0.8, y2=0.9),
            [ScoredLabel(roi_label)],
            ID("rectangle_roi"),
        )
        assert np.array_equal(dataset_item.roi_numpy(rectangle_roi), media.numpy[1:9, 3:13])
        # Checking array returned by "roi_numpy" method with specified Ellipse-shape "roi" parameter
        ellipse_roi = Annotation(
            Ellipse(x1=0.1, y1=0.0, x2=0.9, y2=0.8),
            [ScoredLabel(roi_label)],
            ID("ellipse_roi"),
        )
        assert np.array_equal(dataset_item.roi_numpy(ellipse_roi), media.numpy[0:8, 2:14])
        # Checking array returned by "roi_numpy" method with specified Polygon-shape "roi" parameter
        polygon_roi = Annotation(
            shape=Polygon(
                [
                    Point(0.3, 0.4),
                    Point(0.3, 0.7),
                    Point(0.5, 0.75),
                    Point(0.8, 0.7),
                    Point(0.8, 0.4),
                ]
            ),
            labels=[],
            id=ID("polygon_roi"),
        )
        assert np.array_equal(dataset_item.roi_numpy(polygon_roi), media.numpy[4:8, 5:13])
        # Checking array returned by "roi_numpy" method with not specified "roi" parameter for DatasetItemEntity with
        # "roi" attribute
        roi_specified_dataset_item = DatasetItemEntity(media, annotation_scene, DatasetItemParameters().roi())
        roi_specified_dataset_item.roi_numpy()
        assert np.array_equal(roi_specified_dataset_item.roi_numpy(), media.numpy[1:9, 2:14])

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_item_numpy(self):
        """
        <b>Description:</b>
        Check DatasetItemEntity class "numpy" property

        <b>Input data:</b>
        DatasetItemEntity class object with specified "media", "annotation_scene", "roi", "metadata" and "subset"
        parameters

        <b>Expected results:</b>
        Test passes if array returned by "numpy" property is equal to array returned by "roi_numpy" method

        <b>Steps</b>
        1. Check array returned by "numpy" property for DatasetItemEntity with "roi" attribute is "None"
        2. Check array returned by "numpy" property for DatasetItemEntity with specified "roi" attribute
        """
        # Checking array returned by numpy property for DatasetItemEntity with "roi" attribute is "None"
        none_roi_dataset_item = DatasetItemParameters().default_values_dataset_item()
        assert np.array_equal(none_roi_dataset_item.numpy, none_roi_dataset_item.roi_numpy())
        # Checking array returned by numpy property for DatasetItemEntity with specified "roi" attribute
        roi_specified_dataset_item = DatasetItemParameters().dataset_item()
        assert np.array_equal(roi_specified_dataset_item.numpy, roi_specified_dataset_item.roi_numpy())

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_item_width(self):
        """
        <b>Description:</b>
        Check DatasetItemEntity class "width" property

        <b>Input data:</b>
        DatasetItemEntity class object with specified "media", "annotation_scene", "roi", "metadata" and "subset"
        parameters

        <b>Expected results:</b>
        Test passes if value returned by "width" property is equal to expected

        <b>Steps</b>
        1. Check value returned by "width" property for DatasetItemEntity with "roi" attribute is "None"
        2. Check value returned by "width" property for DatasetItemEntity with specified "roi" attribute
        """
        # Checking value returned by "width" property for DatasetItemEntity with "roi" attribute is "None"
        none_roi_dataset_item = DatasetItemParameters().default_values_dataset_item()
        assert none_roi_dataset_item.width == 16
        # Checking value returned by "width" property for DatasetItemEntity with specified "roi" attribute
        roi_specified_dataset_item = DatasetItemParameters().dataset_item()
        assert roi_specified_dataset_item.width == 12

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_item_height(self):
        """
        <b>Description:</b>
        Check DatasetItemEntity class "height" property

        <b>Input data:</b>
        DatasetItemEntity class object with specified "media", "annotation_scene", "roi", "metadata" and "subset"
        parameters

        <b>Expected results:</b>
        Test passes if value returned by "height" property is equal to expected

        <b>Steps</b>
        1. Check value returned by "height" property for DatasetItemEntity with "roi" attribute is "None"
        2. Check value returned by "height" property for DatasetItemEntity with specified "roi" attribute
        """
        # Checking value returned by "width" property for DatasetItemEntity with None "roi" attribute
        none_roi_dataset_item = DatasetItemParameters().default_values_dataset_item()
        assert none_roi_dataset_item.height == 10
        # Checking value returned by "width" property for DatasetItemEntity with specified "roi" attribute
        roi_specified_dataset_item = DatasetItemParameters().dataset_item()
        assert roi_specified_dataset_item.height == 8

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    @pytest.mark.xfail
    # TODO: Fix this - https://jira.devtools.intel.com/browse/CVS-91526
    def test_dataset_item_get_annotations(self):
        """
        <b>Description:</b>
        Check DatasetItemEntity class "get_annotations" method

        <b>Input data:</b>
        DatasetItemEntity class object with specified "media", "annotation_scene", "roi", "metadata" and "subset"
        parameters

        <b>Expected results:</b>
        Test passes if list returned by "get_annotations" method is equal to expected

        <b>Steps</b>
        1. Check that get_annotations returns all annotations in the dataset item if the ROI is a full box
        2. Check that after adding the parameter "labels", only the annotations with that label are returned
        3. Check that for a ROI that includes only one of the annotations, only that annotation is returned
        """
        # Check that get_annotations returns all items if the ROI is a full box.
        full_box_roi_dataset_item = DatasetItemParameters().default_values_dataset_item()
        full_box_annotations = list(full_box_roi_dataset_item.annotation_scene.annotations)
        result_annotations = full_box_roi_dataset_item.get_annotations(include_empty=True)
        expected_annotations = full_box_annotations
        self.compare_denormalized_annotations(result_annotations, expected_annotations)

        # Check that get_annotations returns only the items with the right label if the "labels" param is used
        first_annotation = full_box_roi_dataset_item.annotation_scene.annotations[0]
        first_annotation_label = first_annotation.get_labels()[0].label
        result_annotations = full_box_roi_dataset_item.get_annotations(
            labels=[first_annotation_label], include_empty=True
        )
        expected_annotations = [first_annotation]
        self.compare_denormalized_annotations(result_annotations, expected_annotations)

        # Check that get_annotations only returns the annotations whose center falls within the ROI
        partial_box_dataset_item = deepcopy(full_box_roi_dataset_item)
        partial_box_dataset_item.roi = Annotation(shape=Rectangle(x1=0.0, y1=0.0, x2=0.4, y2=0.5), labels=[])
        expected_annotation = deepcopy(first_annotation)
        expected_annotation.shape = expected_annotation.shape.denormalize_wrt_roi_shape(
            roi_shape=partial_box_dataset_item.roi.shape
        )
        result_annotations = partial_box_dataset_item.get_annotations(include_empty=True)
        self.compare_denormalized_annotations(result_annotations, [expected_annotation])

        # Check if ignored labels are properly removed
        ignore_labels_dataset_item = DatasetItemParameters().default_values_dataset_item()
        ignore_labels_dataset_item.ignored_labels = ignore_labels_dataset_item.get_shapes_labels(
            include_ignored=True, include_empty=True
        )
        assert ignore_labels_dataset_item.get_annotations(include_empty=True) == []

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_item_append_annotations(self):
        """
        <b>Description:</b>
        Check DatasetItemEntity class "append_annotations" method

        <b>Input data:</b>
        DatasetItemEntity class object with specified "media", "annotation_scene", "roi", "metadata" and "subset"
        parameters

        <b>Expected results:</b>
        Test passes if annotations list returned after "append_annotations" method is equal to expected

        <b>Steps</b>
        1. Check annotations list returned after "append_annotations" method with specified non-included annotations
        2. Check annotations list returned after "append_annotations" method with incorrect shape annotation
        """
        # Checking annotations list returned after "append_annotations" method with specified non-included annotations
        dataset_item = DatasetItemParameters().default_values_dataset_item()
        full_box_annotations = list(dataset_item.annotation_scene.annotations)
        annotations_to_add = self.annotations_to_add()
        normalized_annotations = []
        for annotation in annotations_to_add:
            normalized_annotations.append(
                Annotation(
                    shape=annotation.shape.normalize_wrt_roi_shape(dataset_item.roi.shape),
                    labels=annotation.get_labels(),
                )
            )
        dataset_item.append_annotations(annotations_to_add)
        # Random id is generated for normalized annotations
        normalized_annotations[0].id_ = dataset_item.annotation_scene.annotations[2].id_
        normalized_annotations[1].id_ = dataset_item.annotation_scene.annotations[3].id_
        assert dataset_item.annotation_scene.annotations == full_box_annotations + normalized_annotations
        # Checking annotations list returned after "append_annotations" method with incorrect shape annotation
        incorrect_shape_label = LabelEntity(
            name="Label for incorrect shape",
            domain=Domain.CLASSIFICATION,
            color=Color(red=80, green=70, blue=155),
            id=ID("incorrect_shape_label"),
        )
        incorrect_polygon = Polygon([Point(x=0.01, y=0.1), Point(x=0.35, y=0.1), Point(x=0.35, y=0.1)])
        incorrect_shape_annotation = Annotation(
            shape=incorrect_polygon,
            labels=[ScoredLabel(incorrect_shape_label)],
            id=ID("incorrect_shape_annotation"),
        )
        dataset_item.append_annotations([incorrect_shape_annotation])
        assert dataset_item.annotation_scene.annotations == full_box_annotations + normalized_annotations

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_item_get_roi_labels(self):
        """
        <b>Description:</b>
        Check DatasetItemEntity class "get_roi_labels" method

        <b>Input data:</b>
        DatasetItemEntity class object with specified "media", "annotation_scene", "roi", "metadata" and "subset"
        parameters

        <b>Expected results:</b>
        Test passes if annotations list returned by "get_roi_labels" method is equal to expected

        <b>Steps</b>
        1. Check annotations list returned by "get_roi_labels" for non-specified "labels" parameter
        2. Check annotations list returned by "get_roi_labels" for specified "labels" parameter
        3. Check annotations list returned by "get_roi_labels" if dataset item ignores a label
        """
        dataset_item = DatasetItemParameters().dataset_item()
        roi_labels = DatasetItemParameters.roi_labels()
        # Checking annotations list returned by "get_roi_labels" method with non-specified labels parameter
        # Scenario for "include_empty" is "False"
        assert dataset_item.get_roi_labels() == [roi_labels[0]]
        # Scenario for "include_empty" is "True"
        assert dataset_item.get_roi_labels(include_empty=True) == roi_labels
        # Checking annotations list returned by "get_roi_labels" method with specified labels parameter
        empty_roi_label = roi_labels[1]
        # Scenario for "include_empty" is "False"
        assert dataset_item.get_roi_labels(labels=[empty_roi_label]) == []
        # Scenario for "include_empty" is "True"
        assert dataset_item.get_roi_labels([empty_roi_label], True) == [empty_roi_label]
        # Scenario for ignored labels
        dataset_item.ignored_labels = [empty_roi_label]
        assert dataset_item.get_roi_labels([empty_roi_label], True) == []

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_item_get_shapes_labels(self):
        """
        <b>Description:</b>
        Check DatasetItemEntity class "get_shapes_labels" method

        <b>Input data:</b>
        DatasetItemEntity class object with specified "media", "annotation_scene", "roi", "metadata" and "subset"
        parameters

        <b>Expected results:</b>
        Test passes if labels list returned by "get_shapes_labels" method is equal to expected

        <b>Steps</b>
        1. Check labels list returned by "get_shapes_labels" for non-specified "labels" parameter
        2. Check labels list returned by "get_shapes_labels" for specified "labels" parameter
        3. Check labels list returned by "get_shapes_labels" if dataset_item ignores labels
        """
        dataset_item = DatasetItemParameters().default_values_dataset_item()
        labels = DatasetItemParameters.labels()
        detection_label = labels[0]
        segmentation_label = labels[1]
        # Checking labels list returned by "get_shapes_labels" method with non-specified "labels" parameter
        # Scenario for "include_empty" is "False"
        assert dataset_item.get_shapes_labels() == [detection_label]
        # Scenario for "include_empty" is "True"
        shapes_labels_actual = dataset_item.get_shapes_labels(include_empty=True)
        assert len(shapes_labels_actual) == 2
        assert isinstance(shapes_labels_actual, list)
        assert detection_label in shapes_labels_actual
        assert segmentation_label in shapes_labels_actual
        # Checking labels list returned by "get_shapes_labels" method with specified "labels" parameter
        # Scenario for "include_empty" is "False"
        non_included_label = LabelEntity("Non-included label", Domain.CLASSIFICATION)
        list_labels = [segmentation_label, non_included_label]
        assert dataset_item.get_shapes_labels(labels=list_labels) == []
        # Scenario for "include_empty" is "True", expected that non_included label will not be shown
        assert dataset_item.get_shapes_labels(list_labels, include_empty=True) == [segmentation_label]
        # Check ignore labels functionality
        dataset_item.ignored_labels = [detection_label]
        assert dataset_item.get_shapes_labels(include_empty=True, include_ignored=False) == [segmentation_label]
        assert dataset_item.get_shapes_labels(include_empty=False, include_ignored=True) == [detection_label]

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_item_append_labels(self):
        """
        <b>Description:</b>
        Check DatasetItemEntity class "append_labels" method

        <b>Input data:</b>
        DatasetItemEntity class object with specified "media", "annotation_scene", "roi", "metadata" and "subset"
        parameters

        <b>Expected results:</b>
        Test passes if annotations list returned after using "append_labels" method is equal to expected

        <b>Steps</b>
        1. Check annotations list after "append_labels" method for DatasetItemEntity object with ROI-annotation
        specified in annotation_scene.annotations
        2. Check annotations list after "append_labels" method for DatasetItemEntity object with non-specified
        ROI-annotation in annotation_scene.annotations
        """
        annotation_labels = DatasetItemParameters.labels()
        labels_to_add = self.labels_to_add()
        scored_labels_to_add = [
            ScoredLabel(labels_to_add[0]),
            ScoredLabel(labels_to_add[1]),
        ]
        media = DatasetItemParameters.generate_random_image()
        roi_labels = DatasetItemParameters.roi_labels()
        roi_scored_labels = DatasetItemParameters().roi_scored_labels()
        roi = DatasetItemParameters().roi()
        equal_roi = DatasetItemParameters().roi()
        annotations = DatasetItemParameters().annotations()
        annotations_with_roi = annotations + [equal_roi]
        annotations_scene = AnnotationSceneEntity(annotations_with_roi, AnnotationSceneKind.ANNOTATION)
        # Scenario for checking "append_labels" method for DatasetItemEntity object with ROI-annotation specified in
        # annotation_scene.annotations object
        roi_label_dataset_item = DatasetItemEntity(media, annotations_scene, roi)
        roi_label_dataset_item.append_labels(scored_labels_to_add)
        # Check for include_empty is "False"
        expected_labels = [annotation_labels[0], roi_labels[0], labels_to_add[0]]
        assert roi_label_dataset_item.annotation_scene.get_labels() == expected_labels
        expected_labels = [roi_scored_labels[0], scored_labels_to_add[0]]
        self.check_roi_equal_annotation(roi_label_dataset_item, expected_labels)
        # Check for include_empty is "True"
        expected_labels = annotation_labels + roi_labels + labels_to_add
        assert roi_label_dataset_item.annotation_scene.get_labels(True) == expected_labels
        expected_labels = roi_scored_labels + scored_labels_to_add
        self.check_roi_equal_annotation(roi_label_dataset_item, expected_labels, True)
        # Scenario for checking "append_labels" method for DatasetItemEntity object with non-specified ROI-annotation in
        # annotation_scene.annotations object
        non_roi_dataset_item = DatasetItemParameters().dataset_item()
        non_roi_dataset_item.append_labels(scored_labels_to_add)
        # Check for "include_empty" is "False"
        expected_labels = [annotation_labels[0], roi_labels[0], labels_to_add[0]]
        assert non_roi_dataset_item.annotation_scene.get_labels() == expected_labels
        expected_labels = [roi_scored_labels[0], scored_labels_to_add[0]]
        self.check_roi_equal_annotation(non_roi_dataset_item, expected_labels)
        # Check for "include_empty" is "True"
        expected_labels = annotation_labels + roi_labels + labels_to_add
        assert non_roi_dataset_item.annotation_scene.get_labels(True) == expected_labels
        expected_labels = roi_scored_labels + scored_labels_to_add
        self.check_roi_equal_annotation(non_roi_dataset_item, expected_labels, True)
        # Scenario for "labels" parameter is equal to []
        dataset_item = DatasetItemParameters().dataset_item()
        dataset_item.append_labels([])
        assert dataset_item.annotation_scene.get_labels() == [annotation_labels[0]]
        assert dataset_item.annotation_scene.get_labels(include_empty=True) == annotation_labels

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_item_eq(self):
        """
        <b>Description:</b>
        Check DatasetItemEntity class __eq__ method

        <b>Input data:</b>
        DatasetItemEntity class objects with specified "media", "annotation_scene", "roi", "metadata", "subset"
        and "ignored_labels" parameters

        <b>Expected results:</b>
        Test passes if value returned by __eq__ method is equal to expected

        <b>Steps</b>
        1. Check value returned by __eq__ method for equal DatasetItemEntity objects
        2. Check value returned by __eq__ method for DatasetItemEntity objects with unequal "media", "annotation_scene",
        "roi", "subset" or "ignored_labels" parameters
        3. Check value returned by __eq__ method for DatasetItemEntity objects with unequal "metadata" parameters
        4. Check value returned by __eq__ method for DatasetItemEntity object compared to different type object
        """
        media = DatasetItemParameters.generate_random_image()
        annotation_scene = DatasetItemParameters().annotations_entity()
        roi = DatasetItemParameters().roi()
        metadata = DatasetItemParameters.metadata()
        ignored_labels = DatasetItemParameters.labels()[:1]
        dataset_parameters = {
            "media": media,
            "annotation_scene": annotation_scene,
            "roi": roi,
            "metadata": metadata,
            "subset": Subset.TESTING,
            "ignored_labels": ignored_labels,
        }
        dataset_item = DatasetItemEntity(**dataset_parameters)
        # Checking value returned by __eq__ method for equal DatasetItemEntity objects
        equal_dataset_item = DatasetItemEntity(**dataset_parameters)
        assert dataset_item == equal_dataset_item
        # Checking inequality of DatasetItemEntity objects with unequal initialization parameters
        unequal_annotation_scene = DatasetItemParameters().annotations_entity()
        unequal_annotation_scene.annotations.pop(0)
        unequal_ignored_labels = DatasetItemParameters.labels()[1:]
        unequal_values = [
            ("media", DatasetItemParameters.generate_random_image()),
            ("annotation_scene", unequal_annotation_scene),
            ("roi", None),
            ("subset", Subset.VALIDATION),
            ("ignored_labels", unequal_ignored_labels),
        ]
        for key, value in unequal_values:
            unequal_parameters = dict(dataset_parameters)
            unequal_parameters[key] = value
            unequal_dataset_item = DatasetItemEntity(**unequal_parameters)
            assert dataset_item != unequal_dataset_item, (
                f"Expected False returned for DatasetItemEntity objects with " f"unequal {key} parameters"
            )
        # Checking value returned by __eq__ method for DatasetItemEntity objects with unequal "metadata" parameters
        # expected equality
        unequal_metadata_parameters = dict(dataset_parameters)
        unequal_metadata_parameters["metadata"] = None
        unequal_metadata_dataset_item = DatasetItemEntity(**unequal_metadata_parameters)
        assert dataset_item == unequal_metadata_dataset_item
        # Checking value returned by __eq__ method for DatasetItemEntity object compared to different type object
        assert not dataset_item == str

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_item_deepcopy(self):
        """
        <b>Description:</b>
        Check DatasetItemEntity class __deepcopy__ method

        <b>Input data:</b>
        DatasetItemEntity class objects with specified "media", "annotation_scene", "roi", "metadata" and "subset"
        parameters

        <b>Expected results:</b>
        Test passes if DatasetItemEntity object created by __deepcopy__ method is equal to expected
        """
        dataset_item = DatasetItemParameters().dataset_item()
        copy_dataset = deepcopy(dataset_item)
        assert dataset_item._DatasetItemEntity__roi_lock != copy_dataset._DatasetItemEntity__roi_lock
        assert np.array_equal(dataset_item.media.numpy, copy_dataset.media.numpy)
        assert dataset_item.annotation_scene.annotations == copy_dataset.annotation_scene.annotations
        assert dataset_item.annotation_scene.creation_date == copy_dataset.annotation_scene.creation_date
        assert dataset_item.annotation_scene.editor_name == copy_dataset.annotation_scene.editor_name
        assert dataset_item.annotation_scene.id_ == copy_dataset.annotation_scene.id_
        assert dataset_item.annotation_scene.kind == copy_dataset.annotation_scene.kind
        assert dataset_item.annotation_scene.shapes == copy_dataset.annotation_scene.shapes
        assert dataset_item.roi == copy_dataset.roi
        assert dataset_item.get_metadata() == copy_dataset.get_metadata()
        assert dataset_item.subset == copy_dataset.subset

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_item_append_metadata_item(self):
        """
        <b>Description:</b>
        Check DatasetItemEntity class "append_metadata_item" method

        <b>Input data:</b>
        DatasetItemEntity class object with specified "media", "annotation_scene", "roi", "metadata" and "subset"
        parameters

        <b>Expected results:</b>
        Test passes if "metadata" attribute after "append_metadata_item" method is equal to expected

        <b>Steps</b>
        1. Check "metadata" attribute after "append_metadata_item" method with non-specified "model" parameter
        2. Check "metadata" attribute after "append_metadata_item" method with specified "model" parameter
        """
        dataset_item = DatasetItemParameters().dataset_item()
        expected_metadata = dataset_item.get_metadata()
        # Checking metadata attribute returned after "append_metadata_item" method with non-specified "model" parameter
        data_to_append = TensorEntity(
            name="appended_metadata",
            numpy=np.random.uniform(low=0.0, high=255.0, size=(10, 15, 3)),
        )
        expected_metadata.append(MetadataItemEntity(data=data_to_append))
        dataset_item.append_metadata_item(data=data_to_append)
        assert dataset_item.get_metadata() == expected_metadata
        # Checking metadata attribute returned after "append_metadata_item" method with specified "model" parameter
        metadata_item_with_model = self.metadata_item_with_model()
        data_to_append = metadata_item_with_model.data
        model_to_append = metadata_item_with_model.model
        new_metadata_item_with_model = MetadataItemEntity(data_to_append, model_to_append)
        expected_metadata.append(new_metadata_item_with_model)
        dataset_item.append_metadata_item(data_to_append, model_to_append)
        assert dataset_item.get_metadata() == expected_metadata

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_item_get_metadata_by_name_and_model(self):
        """
        <b>Description:</b>
        Check DatasetItemEntity class "get_metadata_by_name_and_model" method

        <b>Input data:</b>
        DatasetItemEntity class object with specified "media", "annotation_scene", "roi", "metadata" and "subset"
        parameters

        <b>Expected results:</b>
        Test passes if MetadataItemEntity object returned by "get_metadata_by_name_and_model" is equal to expected

        <b>Steps</b>
        1. Check value returned by "get_metadata_by_name_and_model" method for searching metadata object with "model"
        is "None"
        2. Check value returned by "get_metadata_by_name_and_model" method for searching metadata object with specified
        "model" attribute
        3. Check value returned by "get_metadata_by_name_and_model" method for searching non-existing metadata object
        """
        dataset_item = DatasetItemParameters().dataset_item()
        metadata_item_with_model = self.metadata_item_with_model()
        dataset_model = metadata_item_with_model.model
        dataset_item.append_metadata_item(metadata_item_with_model.data, dataset_model)
        dataset_metadata = dataset_item.get_metadata()
        # Checking "get_metadata_by_name_and_model" method for "model" parameter is "None"
        assert dataset_item.get_metadata_by_name_and_model("test_metadata", None) == [dataset_metadata[0]]
        # Checking "get_metadata_by_name_and_model" method for specified "model" parameter
        assert dataset_item.get_metadata_by_name_and_model("appended_metadata_with_model", dataset_model) == [
            dataset_metadata[2]
        ]
        # Checking "get_metadata_by_name_and_model" method for searching non-existing metadata
        assert dataset_item.get_metadata_by_name_and_model("test_metadata", dataset_model) == []
        assert dataset_item.get_metadata_by_name_and_model("appended_metadata_with_model", None) == []

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_item_setters(self):
        """
        <b>Description:</b>
        Check DatasetItemEntity class "roi", "subset" and "annotation_scene" setters

        <b>Input data:</b>
        DatasetItemEntity class object with specified "media", "annotation_scene", "roi", "metadata" and "subset"
        parameters

        <b>Expected results:</b>
        Test passes if assigned values of "roi", "subset" and "annotation_scene" properties are equal to expected

        <b>Steps</b>
        1. Check value returned by "roi" property after using @roi.setter
        2. Check value returned by "subset" property after using @subset.setter
        3. Check value returned by "annotation_scene" property after using @subset.annotation_scene
        """
        dataset_item = DatasetItemParameters().dataset_item()
        # Checking value returned by "roi" property after using @roi.setter
        new_roi_label = ScoredLabel(LabelEntity("new ROI label", Domain.DETECTION))
        new_dataset_roi = Annotation(Rectangle(x1=0.2, y1=0.2, x2=1.0, y2=1.0), [new_roi_label])
        dataset_item.roi = new_dataset_roi
        assert dataset_item.roi == new_dataset_roi
        # Checking value returned by subset property after using @subset.setter
        new_subset = Subset.TRAINING
        dataset_item.subset = new_subset
        assert dataset_item.subset == new_subset
        # Checking value returned by annotation_scene property after using @annotation_scene.setter
        new_annotation_label = ScoredLabel(LabelEntity("new annotation label", Domain.CLASSIFICATION))
        new_annotation = Annotation(Rectangle(x1=0.1, y1=0, x2=0.9, y2=1.0), [new_annotation_label])
        new_annotation_scene = AnnotationSceneEntity([new_annotation], AnnotationSceneKind.PREDICTION)
        dataset_item.annotation_scene = new_annotation_scene
        assert dataset_item.annotation_scene == new_annotation_scene

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    @pytest.mark.parametrize("func_name", ["dataset_item", "dataset_item_with_id"])
    def test_wrap(self, func_name):
        constructor = DatasetItemParameters()
        func = getattr(constructor, func_name)
        item: DatasetItemEntity = func()

        new_media = DatasetItemParameters().generate_random_image()
        assert item.media != new_media

        new_subset = Subset.PSEUDOLABELED
        assert item.subset != new_subset

        new_metadata = DatasetItemParameters().metadata()
        assert item.get_metadata() != new_metadata

        new_item = item.wrap(media=new_media, subset=new_subset, metadata=new_metadata)
        assert new_item.media == new_media
        assert new_item.subset == new_subset
        assert new_item.get_metadata() == new_metadata

        if hasattr(item, "id_"):
            new_id = ID("new_id")
            assert item.id_ != new_id
            item = item.wrap(id_=new_id)
            assert item.id_ == new_id



"""Parser modules for otx.core.ov.graph.parsers."""
# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from typing import List


def type_parser(graph, types) -> List[str]:
    """Type Parser from graph, types."""
    found = []
    for node in graph:
        if node.type in types:
            found.append(node.name)
    return found


def result_parser(graph) -> List[str]:
    """Result Parser from graph."""
    return type_parser(graph, ["Result"])


def parameter_parser(graph) -> List[str]:
    """Parameter Parser from graph."""
    return type_parser(graph, ["Parameter"])



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-Lggggcense-Identifier: Apache-2.0
#

import pytest
import torch

from otx.core.ov.ops.movements import (
    BroadcastV3,
    ConcatV0,
    GatherV0,
    GatherV1,
    PadV1,
    ScatterNDUpdateV3,
    ScatterUpdateV3,
    ShuffleChannelsV0,
    SplitV1,
    StridedSliceV1,
    TileV0,
    TransposeV1,
    VariadicSplitV1,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestPadV1:
    @e2e_pytest_unit
    def test_invalid_attr(self):
        with pytest.raises(ValueError):
            PadV1("dummy", shape=(1,), pad_mode="error")

    @e2e_pytest_unit
    def test_get_torch_pad_mode(self):
        assert "constant" == PadV1.get_torch_pad_mode("constant")
        assert "replicate" == PadV1.get_torch_pad_mode("edge")
        assert "reflect" == PadV1.get_torch_pad_mode("reflect")
        with pytest.raises(NotImplementedError):
            PadV1.get_torch_pad_mode("symmetric")
        with pytest.raises(NotImplementedError):
            PadV1.get_torch_pad_mode("error")

    @e2e_pytest_unit
    def test_get_torch_pad_dim(self):
        assert [10, 10, 10, 10] == PadV1.get_torch_pad_dim([10, 10], [10, 10])
        assert [9, 4, 8, 3, 7, 2, 6, 1] == PadV1.get_torch_pad_dim(range(6, 10), range(0, 5))

    @e2e_pytest_unit
    def test_forward(self):
        for mode in ("constant", "edge", "reflect"):
            op = PadV1("dummy", shape=(1,), pad_mode=mode)
            input = torch.empty(3, 3, 4, 2)
            output = op(input, [1, 1], [1, 1])
            assert output.shape == (3, 3, 6, 4)

            output = op(input, [2, 2, 1], [1, 2, 1])
            assert output.shape == (3, 6, 8, 4)


class TestConcatV0:
    @e2e_pytest_unit
    def test_forward(self):
        op = ConcatV0("dummy", shape=(1,), axis=1)
        input_1 = torch.rand(4, 4, 32, 32)
        input_2 = torch.rand(4, 8, 32, 32)
        output = op(input_1, input_2)
        ref = torch.cat((input_1, input_2), 1)
        assert torch.equal(output, ref)


class TestTransposeV1:
    @e2e_pytest_unit
    def test_forward(self):
        op = TransposeV1("dummy", shape=(1,))
        input = torch.randn(1, 2, 3, 4, 5)

        assert op(input, torch.tensor([])).shape == (5, 4, 3, 2, 1)
        assert op(input, torch.tensor([0, 2, 1, 4, 3])).shape == (1, 3, 2, 5, 4)


class TestGatherV0:
    @e2e_pytest_unit
    def test_forward(self):
        op = GatherV0("dummy", shape=(1,), batch_dims=0)
        output = op(torch.tensor([1, 2, 3, 4, 5]), torch.tensor([0, 0, 4]), torch.tensor(0))
        ref = torch.tensor([1, 1, 5])
        assert torch.equal(output, ref)

        op = GatherV0("dummy", shape=(1,), batch_dims=0)
        output = op(torch.tensor([1, 2, 3, 4, 5]), torch.tensor(4), torch.tensor(0))
        ref = torch.tensor(5)
        assert torch.equal(output, ref)

        op = GatherV0("dummy", shape=(1,), batch_dims=-1)
        output = op(
            torch.tensor([[1, 2, 3, 4, 5], [6, 7, 8, 9, 10]]),
            torch.tensor([[0, 0, 4], [4, 0, 0]]),
            torch.tensor(1),
        )
        ref = torch.tensor([[1, 1, 5], [10, 6, 6]])
        assert torch.equal(output, ref)

        op = GatherV0("dummy", shape=(1,), batch_dims=1)
        output = op(
            torch.tensor([[1, 2, 3, 4, 5], [6, 7, 8, 9, 10]]),
            torch.tensor([[0, 0, 4], [4, 0, 0]]),
            torch.tensor(1),
        )
        ref = torch.tensor([[1, 1, 5], [10, 6, 6]])
        assert torch.equal(output, ref)

        op = GatherV0("dummy", shape=(1,), batch_dims=2)
        output = op(
            torch.tensor(
                [
                    [[1, 2, 3, 4, 5], [6, 7, 8, 9, 10]],
                    [[11, 12, 13, 14, 15], [16, 17, 18, 19, 20]],
                ]
            ),
            torch.tensor([[[0, 0, 4], [4, 0, 0]], [[1, 2, 4], [4, 3, 2]]]),
            torch.tensor(2),
        )
        ref = torch.tensor([[[1, 1, 5], [10, 6, 6]], [[12, 13, 15], [20, 19, 18]]])
        assert torch.equal(output, ref)

        op = GatherV0("dummy", shape=(1,), batch_dims=1)
        output = op(
            torch.tensor(
                [
                    [
                        [
                            [1, 2, 3, 4],
                            [5, 6, 7, 8],
                            [9, 10, 11, 12],
                            [13, 14, 15, 16],
                            [17, 18, 19, 20],
                        ]
                    ],
                    [
                        [
                            [21, 22, 23, 24],
                            [25, 26, 27, 28],
                            [29, 30, 31, 32],
                            [33, 34, 35, 36],
                            [37, 38, 39, 40],
                        ]
                    ],
                ]
            ),
            torch.tensor([[1, 2, 4], [4, 3, 2]]),
            torch.tensor(2),
        )
        ref = torch.tensor(
            [
                [[[5, 6, 7, 8], [9, 10, 11, 12], [17, 18, 19, 20]]],
                [[[37, 38, 39, 40], [33, 34, 35, 36], [29, 30, 31, 32]]],
            ]
        )
        assert torch.equal(output, ref)


class TestGatherV1:
    @e2e_pytest_unit
    def test_forward(self):
        op = GatherV1("dummy", shape=(1,))
        input = torch.tensor([[1, 2], [3, 4]])
        output = op(input, torch.tensor([[0, 0], [1, 0]]), 1)
        ref = torch.gather(input, 1, torch.tensor([[0, 0], [1, 0]]))
        assert torch.equal(output, ref)


class TestStridedSliceV1:
    @e2e_pytest_unit
    def test_forward(self):
        op = StridedSliceV1(
            "dummy",
            shape=(1,),
            begin_mask=[0, 1, 1],
            end_mask=[1, 1, 0],
            new_axis_mask=[0, 0, 0],
            shrink_axis_mask=[0, 0, 0],
        )
        output = op(
            torch.randn(2, 3, 4),
            torch.tensor([1, 0, 0]),
            torch.tensor([0, 0, 2]),
            torch.tensor([1, 1, 1]),
        )
        assert output.shape == (1, 3, 2)

        op = StridedSliceV1(
            "dummy",
            shape=(1,),
            begin_mask=[0, 1, 1],
            end_mask=[1, 1, 0],
            new_axis_mask=[1, 0, 0],
            shrink_axis_mask=[0, 0, 0],
        )
        output = op(
            torch.randn(2, 3, 4),
            torch.tensor([0, 0, 0]),
            torch.tensor([1, 0, 4]),
            torch.tensor([1, 1, 1]),
        )
        assert output.shape == (1, 2, 3, 4)


class TestSplitV1:
    @e2e_pytest_unit
    def test_forward(self):
        op = SplitV1("dummy", shape=(1,), num_splits=2)
        input = torch.randn(4, 8, 32)
        outputs = op(input, 1)
        refs = torch.split(input, 4, 1)
        for output, ref in zip(outputs, refs):
            assert torch.equal(output, ref)


class TestVariadicSplitV1:
    @e2e_pytest_unit
    def test_forward(self):
        op = VariadicSplitV1("dummy", shape=(1,))
        outputs = op(torch.randn(6, 12, 10, 24), 0, torch.tensor([1, 2, 3]))
        ref_shapes = ((1, 12, 10, 24), (2, 12, 10, 24), (3, 12, 10, 24))
        for output, ref_shape in zip(outputs, ref_shapes):
            assert output.shape == ref_shape

        outputs = op(torch.randn(6, 12, 10, 24), 0, torch.tensor([-1, 2]))
        ref_shapes = ((4, 12, 10, 24), (2, 12, 10, 24))
        for output, ref_shape in zip(outputs, ref_shapes):
            assert output.shape == ref_shape


class TestShuffleChannelsV0:
    @e2e_pytest_unit
    def test_forward(self):
        op = ShuffleChannelsV0("dummy", shape=(1,))
        input = torch.randn(4, 8, 32, 32)
        output = op(input)
        assert torch.equal(output, input)

        op = ShuffleChannelsV0("dummy", shape=(1,), group=2)
        input = torch.randn(4, 8, 32, 32)
        output = op(input)
        assert not torch.equal(output, input)


class TestBroadcastV3:
    @e2e_pytest_unit
    def test_invalid_attr(self):
        with pytest.raises(ValueError):
            BroadcastV3("dummy", shape=(1,), mode="error")

    @e2e_pytest_unit
    def test_forward(self):
        op = BroadcastV3("dummy", shape=(1,), mode="numpy")
        output = op(torch.randn(16, 1, 1), torch.tensor([1, 16, 50, 50]))
        assert output.shape == (1, 16, 50, 50)

        op = BroadcastV3("dummy", shape=(1,), mode="explicit")
        output = op(torch.randn(16), torch.tensor([1, 16, 50, 50]), torch.tensor([1]))
        assert output.shape == (1, 16, 50, 50)


class TestScatterNDUpdateV3:
    @e2e_pytest_unit
    def test_forward(self):
        ScatterNDUpdateV3("dummy", shape=(1,))
        # TODO
        pass


class TestScatterUpdateV3:
    @e2e_pytest_unit
    def test_forward(self):
        op = ScatterUpdateV3("dummy", shape=(1,))
        updates = torch.arange(1, 11).reshape((2, 5))
        input = torch.zeros(3, 5, dtype=updates.dtype)
        indices = torch.tensor([[0, 1, 2, 0]])

        output = op(input, indices, updates, torch.tensor(0))
        ref = torch.tensor([[1, 0, 0, 4, 0], [0, 2, 0, 0, 0], [0, 0, 3, 0, 0]])
        assert torch.equal(output, ref)


class TestTileV0:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 3, 256, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_forward(self):
        op = TileV0("dummy", shape=(self.shape,))
        output = op(self.input, torch.tensor([2, 2, 2, 2]))
        assert torch.equal(output, torch.tile(self.input, [2, 2, 2, 2]))



from unittest.mock import MagicMock, patch

import openvino_telemetry as ovtm
import pytest

from otx import __version__
from otx.cli.utils.telemetry import (
    close_telemetry_session,
    init_telemetry_session,
    send_cmd_results,
    send_version,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestTelemetry:
    # FIXME: is there a way to get private constants for the testing?
    __TM_CATEGORY_OTX = "otx"
    __TM_MEASUREMENT_ID = "UA-17808594-29"
    __TM_ACTION_VERSION = "version"
    __TM_ACTION_CMD_SUCCESS = "success"
    __TM_ACTION_CMD_FAILURE = "failure"
    __TM_ACTION_CMD_EXCEPTION = "exception"
    __TM_ACTION_ERROR = "error"

    @e2e_pytest_unit
    @patch("otx.cli.utils.telemetry.tm")
    @patch("otx.cli.utils.telemetry.send_version")
    def test_init_telemetry_session(
        self,
        mock_send_version,
        mock_ovtm,
    ):
        mock_telemetry = MagicMock()
        mock_telemetry.start_session = MagicMock()
        mock_ovtm.Telemetry = MagicMock(return_value=mock_telemetry)

        init_telemetry_session()

        mock_ovtm.Telemetry.assert_called_once_with(
            app_name=self.__TM_CATEGORY_OTX, app_version=str(__version__), tid=self.__TM_MEASUREMENT_ID
        )
        mock_telemetry.start_session.assert_called_once_with(self.__TM_CATEGORY_OTX)
        mock_send_version.assert_called_once_with(mock_telemetry)

    @e2e_pytest_unit
    def test_close_telemetry_session(self):
        mock_ovtm_instance = MagicMock(spec=ovtm.Telemetry)

        close_telemetry_session(mock_ovtm_instance)

        mock_ovtm_instance.end_session.assert_called_once_with(self.__TM_CATEGORY_OTX)
        mock_ovtm_instance.force_shutdown.assert_called_once_with(1.0)

        with pytest.raises(RuntimeError):
            close_telemetry_session(0)

    @e2e_pytest_unit
    @patch("otx.cli.utils.telemetry.__send_event")
    def test_send_version(self, mock_send_event):
        mock_ovtm_instance = MagicMock(spec=ovtm.Telemetry)
        send_version(mock_ovtm_instance)
        mock_send_event.assert_called_once_with(mock_ovtm_instance, self.__TM_ACTION_VERSION, str(__version__))

        with pytest.raises(RuntimeError):
            send_version(None)

    @e2e_pytest_unit
    @patch("otx.cli.utils.telemetry.__send_event")
    @patch("otx.cli.utils.telemetry.__send_error")
    def test_send_cmd_results(
        self,
        mock_send_error,
        mock_send_event,
    ):
        with pytest.raises(RuntimeError):
            send_cmd_results(None, "something", None)

        mock_ovtm_instance = MagicMock(spec=ovtm.Telemetry)

        # with invalid results arg
        results = None

        with pytest.raises(RuntimeError):
            send_cmd_results(mock_ovtm_instance, "something", results)

        mock_send_error.reset_mock()
        mock_send_event.reset_mock()

        # with empty dict
        results = {}
        send_cmd_results(mock_ovtm_instance, "something", results)
        mock_send_error.assert_called_once()
        mock_send_event.assert_not_called()

        mock_send_error.reset_mock()
        mock_send_event.reset_mock()

        # with failure retcode
        results = {"retcode": 1, "some": "results"}
        cmd = "cmd"
        send_cmd_results(mock_ovtm_instance, cmd, results)
        mock_send_error.assert_not_called()
        mock_send_event.assert_called_once_with(
            mock_ovtm_instance, self.__TM_ACTION_CMD_FAILURE, dict(cmd=cmd, **results)
        )

        mock_send_error.reset_mock()
        mock_send_event.reset_mock()

        # with success retcode
        results = {"retcode": 0, "some": "results"}
        cmd = "cmd"
        send_cmd_results(mock_ovtm_instance, cmd, results)
        mock_send_error.assert_not_called()
        mock_send_event.assert_called_once_with(
            mock_ovtm_instance, self.__TM_ACTION_CMD_SUCCESS, dict(cmd=cmd, **results)
        )

        mock_send_error.reset_mock()
        mock_send_event.reset_mock()

        # with exception retcode
        results = {"retcode": -1, "some": "results"}
        cmd = "cmd"
        send_cmd_results(mock_ovtm_instance, cmd, results)
        mock_send_error.assert_not_called()
        mock_send_event.assert_called_once_with(
            mock_ovtm_instance, self.__TM_ACTION_CMD_EXCEPTION, dict(cmd=cmd, **results)
        )

        mock_send_error.reset_mock()
        mock_send_event.reset_mock()



"""Samplers for imbalanced and incremental learning."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# flake8: noqa

from .balanced_sampler import BalancedSampler
from .cls_incr_sampler import ClsIncrSampler

__all__ = ["BalancedSampler", "ClsIncrSampler"]



"""Module for otx.core.ov.models."""
# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from .mmov_model import MMOVModel
from .ov_model import OVModel  # type: ignore[attr-defined]
from .parser_mixin import ParserMixin  # type: ignore[attr-defined]

__all__ = [
    "MMOVModel",
    "OVModel",
    "ParserMixin",
]



"""Utils function for otx.core.ov.ops."""
# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from openvino.pyopenvino import Node  # pylint: disable=no-name-in-module

from .builder import OPS

# TODO: We moved the location of otx.mpa.utils.logger, we need to revert the logger in that code again.


def get_dynamic_shape(output):
    """Getter function for dynamic shape."""
    shape = [str(i) for i in output.get_partial_shape()]
    for i, shape_ in enumerate(shape):
        try:
            shape_ = int(shape_)
        except ValueError:
            shape_ = -1
        shape[i] = shape_
    return shape


def convert_op_to_torch(op_node: Node):
    """Convert op Node to torch."""
    op_type = op_node.get_type_name()
    op_version = op_node.get_version()

    try:
        torch_module = OPS.get_by_type_version(op_type, op_version).from_ov(op_node)
    except Exception as e:
        # logger.error(e)
        # logger.error(op_type)
        # logger.error(op_version)
        # logger.error(op_node.get_attributes())
        raise e

    return torch_module



from pathlib import Path
from tempfile import TemporaryDirectory

import pytest
import yaml

from otx.cli.utils.config import configure_dataset, override_parameters
from tests.test_suite.e2e_test_system import e2e_pytest_unit


@pytest.fixture
def mock_parameters():
    return {"a": {"a.a": {"value": 1, "default_value": 2}}}


@e2e_pytest_unit
def test_override_mock_parameters(mock_parameters):
    overrides = {"a": {"a.a": {"value": 3, "default_value": 4}}}
    override_parameters(overrides, mock_parameters)

    assert mock_parameters["a"]["a.a"]["value"] == 3
    assert mock_parameters["a"]["a.a"]["default_value"] == 4


@e2e_pytest_unit
def test_override_mock_parameters_not_allowed_key(mock_parameters):
    overrides = {"a": {"a.a": {"fake": 3, "default_value": 4}}}

    with pytest.raises(ValueError):
        override_parameters(overrides, mock_parameters)


@e2e_pytest_unit
def test_override_mock_parameters_non_exist_key(mock_parameters):
    overrides = {"a": {"a.b": {"value": 3, "default_value": 4}}}

    with pytest.raises(ValueError):
        override_parameters(overrides, mock_parameters)


@e2e_pytest_unit
def test_configure_dataset(mocker):
    # prepare
    def mock_contain(self, key):
        return key in self.__dict__

    mock_args = mocker.MagicMock()
    mock_args.__contains__ = mock_contain
    mock_args.train_ann_files = "train_ann_files"
    mock_args.train_data_roots = "train_data_roots"
    mock_args.val_ann_files = "val_ann_files"
    mock_args.val_data_roots = "val_data_roots"
    mock_args.unlabeled_file_list = "unlabeled_file_list"
    mock_args.unlabeled_data_roots = "unlabeled_data_roots"
    mock_args.test_ann_files = "test_ann_files"
    mock_args.test_data_roots = "test_data_roots"
    mock_args.data = None
    # run
    data_config = configure_dataset(mock_args)

    # check
    assert data_config["data"]["train"]["ann-files"] == str(Path(mock_args.train_ann_files).absolute())
    assert data_config["data"]["train"]["data-roots"] == str(Path(mock_args.train_data_roots).absolute())
    assert data_config["data"]["val"]["ann-files"] == str(Path(mock_args.val_ann_files).absolute())
    assert data_config["data"]["val"]["data-roots"] == str(Path(mock_args.val_data_roots).absolute())
    assert data_config["data"]["unlabeled"]["file-list"] == str(Path(mock_args.unlabeled_file_list).absolute())
    assert data_config["data"]["unlabeled"]["data-roots"] == str(Path(mock_args.unlabeled_data_roots).absolute())
    assert data_config["data"]["test"]["ann-files"] == str(Path(mock_args.test_ann_files).absolute())
    assert data_config["data"]["test"]["data-roots"] == str(Path(mock_args.test_data_roots).absolute())


@e2e_pytest_unit
def test_configure_dataset_with_data_args(mocker):
    mock_args = mocker.MagicMock()

    with TemporaryDirectory() as tmp_dir:
        data_yaml_path = Path(tmp_dir) / "data.yaml"
        mock_data = {"data": {"train": {"ann-files": "a", "data-roots": "b"}}}
        with open(data_yaml_path, "w") as f:
            yaml.dump(mock_data, f)

        data_config = configure_dataset(mock_args, str(data_yaml_path))

    assert data_config["data"]["train"]["ann-files"] == mock_data["data"]["train"]["ann-files"]
    assert data_config["data"]["train"]["data-roots"] == mock_data["data"]["train"]["data-roots"]



"""OTX CLI."""

# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import os

# How to make an Action Template invisible in Geti
# Check FEATURE_FLAGS_OTX_ACTION_TASKS in the API to determine whether to use the Action
# Always 1 in the OTX CLI
os.environ["FEATURE_FLAGS_OTX_ACTION_TASKS"] = "1"



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-Lggggcense-Identifier: Apache-2.0
#

import torch

from otx.core.ov.ops.reductions import (
    ReduceMeanV1,
    ReduceMinV1,
    ReduceProdV1,
    ReduceSumV1,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestReduceMeanV1:
    @e2e_pytest_unit
    def test_forward(self):
        op = ReduceMeanV1("dummy", shape=(1,), keep_dims=False)
        input = torch.randn(6, 12, 10, 24)

        output = op(input, [])
        assert torch.equal(output, input)

        output = op(input, torch.tensor([1, 2]))
        assert output.shape == (6, 24)
        assert torch.equal(output, torch.mean(input, dim=(1, 2)))

        output = op(input, torch.tensor([-1]))
        assert output.shape == (6, 12, 10)
        assert torch.equal(output, torch.mean(input, dim=-1))

        op = ReduceMeanV1("dummy", shape=(1,), keep_dims=True)

        output = op(input, [])
        assert torch.equal(output, input)

        output = op(input, torch.tensor([1, 2]))
        assert output.shape == (6, 1, 1, 24)
        assert torch.equal(output, torch.mean(input, dim=(1, 2), keepdim=True))

        output = op(input, torch.tensor([-1]))
        assert output.shape == (6, 12, 10, 1)
        assert torch.equal(output, torch.mean(input, dim=-1, keepdim=True))


class TestReduceProdV1:
    @e2e_pytest_unit
    def test_forward(self):
        op = ReduceProdV1("dummy", shape=(1,), keep_dims=False)
        input = torch.randn(6, 12, 10, 24)

        output = op(input, [])
        assert torch.equal(output, input)

        output = op(input, torch.tensor([1, 2]))
        assert output.shape == (6, 24)

        output = op(input, torch.tensor([-1]))
        assert output.shape == (6, 12, 10)

        op = ReduceProdV1("dummy", shape=(1,), keep_dims=True)
        input = torch.randn(6, 12, 10, 24)

        output = op(input, [])
        assert torch.equal(output, input)

        output = op(input, torch.tensor([1, 2]))
        assert output.shape == (6, 1, 1, 24)

        output = op(input, torch.tensor([-1]))
        assert output.shape == (6, 12, 10, 1)


class TestReduceMinV1:
    @e2e_pytest_unit
    def test_forward(self):
        op = ReduceMinV1("dummy", shape=(1,), keep_dims=False)
        input = torch.randn(6, 12, 10, 24)

        output = op(input, [])
        assert torch.equal(output, input)

        output = op(input, torch.tensor([1, 2]))
        assert output.shape == (6, 24)

        output = op(input, torch.tensor([-1]))
        assert output.shape == (6, 12, 10)

        op = ReduceMinV1("dummy", shape=(1,), keep_dims=True)
        input = torch.randn(6, 12, 10, 24)

        output = op(input, [])
        assert torch.equal(output, input)

        output = op(input, torch.tensor([1, 2]))
        assert output.shape == (6, 1, 1, 24)

        output = op(input, torch.tensor([-1]))
        assert output.shape == (6, 12, 10, 1)


class TestReduceSumV1:
    @e2e_pytest_unit
    def test_forward(self):
        op = ReduceSumV1("dummy", shape=(1,), keep_dims=False)
        input = torch.randn(6, 12, 10, 24)

        output = op(input, [])
        assert torch.equal(output, input)

        output = op(input, torch.tensor([1, 2]))
        assert output.shape == (6, 24)
        assert torch.equal(output, torch.sum(input, dim=(1, 2)))

        output = op(input, torch.tensor([-1]))
        assert output.shape == (6, 12, 10)
        assert torch.equal(output, torch.sum(input, dim=-1))

        op = ReduceSumV1("dummy", shape=(1,), keep_dims=True)
        input = torch.randn(6, 12, 10, 24)

        output = op(input, [])
        assert torch.equal(output, input)

        output = op(input, torch.tensor([1, 2]))
        assert output.shape == (6, 1, 1, 24)
        assert torch.equal(output, torch.sum(input, dim=(1, 2), keepdim=True))

        output = op(input, torch.tensor([-1]))
        assert output.shape == (6, 12, 10, 1)
        assert torch.equal(output, torch.sum(input, dim=-1, keepdim=True))



"""Initial file for mmdetection heads."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .cross_dataset_detector_head import CrossDatasetDetectorHead
from .custom_anchor_generator import SSDAnchorGeneratorClustered
from .custom_atss_head import CustomATSSHead
from .custom_retina_head import CustomRetinaHead
from .custom_roi_head import CustomRoIHead
from .custom_ssd_head import CustomSSDHead
from .custom_vfnet_head import CustomVFNetHead
from .custom_yolox_head import CustomYOLOXHead

__all__ = [
    "CrossDatasetDetectorHead",
    "SSDAnchorGeneratorClustered",
    "CustomATSSHead",
    "CustomRetinaHead",
    "CustomSSDHead",
    "CustomRoIHead",
    "CustomVFNetHead",
    "CustomYOLOXHead",
]



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest

from otx.algorithms.detection.utils.utils import (
    ColorPalette,
    generate_label_schema,
    get_det_model_api_configuration,
)
from otx.api.entities.model_template import TaskType, task_type_to_label_domain
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.api.parameters_validation.validation_helper import (
    check_value_error_exception_raised,
)


@e2e_pytest_unit
def test_get_det_model_api_configuration():
    classes = ("rectangle", "ellipse", "triangle")
    label_schema = generate_label_schema(classes, task_type_to_label_domain(TaskType.DETECTION))
    det_thr = 0.5
    model_api_cfg = get_det_model_api_configuration(label_schema, TaskType.DETECTION, det_thr)

    assert len(model_api_cfg) > 0
    assert model_api_cfg[("model_info", "confidence_threshold")] == str(det_thr)



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from otx.algorithms.common.adapters.mmdeploy.utils.operations_domain import (
    DOMAIN_CUSTOM_OPS_NAME,
    add_domain,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


@e2e_pytest_unit
def test_add_domain():
    assert add_domain("abc") == DOMAIN_CUSTOM_OPS_NAME + "::" + "abc"



import json
from copy import deepcopy
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List
from unittest.mock import MagicMock

import pytest

import otx
from otx.api.configuration.helper import create as create_conf_hp
from otx.api.entities.model import ModelEntity
from otx.api.entities.model_template import TaskType
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.cli.registry import find_and_parse_model_template
from otx.cli.utils import hpo
from otx.cli.utils.hpo import (
    HpoCallback,
    HpoDataset,
    HpoRunner,
    TaskEnvironmentManager,
    TaskManager,
    Trainer,
    get_best_hpo_weight,
    run_hpo,
    run_trial,
)
from otx.hpo.hpo_base import TrialStatus
from tests.test_suite.e2e_test_system import e2e_pytest_unit

CLASSIFCATION_TASK = {TaskType.CLASSIFICATION}
DETECTION_TASK = {TaskType.DETECTION, TaskType.INSTANCE_SEGMENTATION, TaskType.ROTATED_DETECTION}
SEGMENTATION_TASK = {TaskType.SEGMENTATION}
ANOMALY_TASK = {TaskType.ANOMALY_CLASSIFICATION, TaskType.ANOMALY_DETECTION, TaskType.ANOMALY_SEGMENTATION}
MMCV_TASK = CLASSIFCATION_TASK | DETECTION_TASK | SEGMENTATION_TASK
ALL_TASK = MMCV_TASK | ANOMALY_TASK
OTX_ROOT_PATH = Path(otx.__file__).parent


class TestTaskManager:
    @e2e_pytest_unit
    @pytest.mark.parametrize("task", MMCV_TASK)
    def test_is_mmcv_framework_task(self, task: TaskType):
        task_manager = TaskManager(task)
        assert task_manager.is_mmcv_framework_task()

    @e2e_pytest_unit
    @pytest.mark.parametrize("task", ANOMALY_TASK)
    def test_is_not_mmcv_framework_task(self, task: TaskType):
        task_manager = TaskManager(task)
        assert not task_manager.is_mmcv_framework_task()

    @e2e_pytest_unit
    @pytest.mark.parametrize("task", CLASSIFCATION_TASK)
    def test_is_cls_framework_task(self, task: TaskType):
        task_manager = TaskManager(task)
        assert task_manager.is_cls_framework_task()

    @e2e_pytest_unit
    @pytest.mark.parametrize("task", ALL_TASK - CLASSIFCATION_TASK)
    def test_is_not_cls_framework_task(self, task: TaskType):
        task_manager = TaskManager(task)
        assert not task_manager.is_cls_framework_task()

    @e2e_pytest_unit
    @pytest.mark.parametrize("task", DETECTION_TASK)
    def test_is_det_framework_task(self, task: TaskType):
        task_manager = TaskManager(task)
        assert task_manager.is_det_framework_task()

    @e2e_pytest_unit
    @pytest.mark.parametrize("task", ALL_TASK - DETECTION_TASK)
    def test_is_not_det_framework_task(self, task: TaskType):
        task_manager = TaskManager(task)
        assert not task_manager.is_det_framework_task()

    @e2e_pytest_unit
    @pytest.mark.parametrize("task", SEGMENTATION_TASK)
    def test_is_seg_framework_task(self, task: TaskType):
        task_manager = TaskManager(task)
        assert task_manager.is_seg_framework_task()

    @e2e_pytest_unit
    @pytest.mark.parametrize("task", ALL_TASK - SEGMENTATION_TASK)
    def test_is_not_seg_framework_task(self, task: TaskType):
        task_manager = TaskManager(task)
        assert not task_manager.is_seg_framework_task()

    @e2e_pytest_unit
    @pytest.mark.parametrize("task", ANOMALY_TASK)
    def test_is_anomaly_framework_task(self, task: TaskType):
        task_manager = TaskManager(task)
        assert task_manager.is_anomaly_framework_task()

    @e2e_pytest_unit
    @pytest.mark.parametrize("task", ALL_TASK - ANOMALY_TASK)
    def test_is_not_anomaly_framework_task(self, task: TaskType):
        task_manager = TaskManager(task)
        assert not task_manager.is_anomaly_framework_task()

    @e2e_pytest_unit
    @pytest.mark.parametrize("task", MMCV_TASK)
    def test_get_mmcv_batch_size_name(self, task: TaskType):
        task_manager = TaskManager(task)
        assert task_manager.get_batch_size_name() == "learning_parameters.batch_size"

    @e2e_pytest_unit
    @pytest.mark.parametrize("task", ANOMALY_TASK)
    def test_get_anomaly_batch_size_name(self, task: TaskType):
        task_manager = TaskManager(task)
        assert task_manager.get_batch_size_name() == "learning_parameters.train_batch_size"

    @e2e_pytest_unit
    def test_get_unknown_task_batch_size_name(self, mocker):
        mock_func1 = mocker.patch.object(TaskManager, "is_mmcv_framework_task")
        mock_func1.return_value = False
        mock_func2 = mocker.patch.object(TaskManager, "is_anomaly_framework_task")
        mock_func2.return_value = False

        task_manager = TaskManager(mocker.MagicMock())

        with pytest.raises(RuntimeError):
            task_manager.get_batch_size_name()

    @e2e_pytest_unit
    @pytest.mark.parametrize("task", MMCV_TASK)
    def test_get_mmcv_epoch_name(self, task: TaskType):
        task_manager = TaskManager(task)
        assert task_manager.get_epoch_name() == "num_iters"

    @e2e_pytest_unit
    @pytest.mark.parametrize("task", ANOMALY_TASK)
    def test_get_anomaly_epoch_name(self, task: TaskType):
        task_manager = TaskManager(task)
        assert task_manager.get_epoch_name() == "max_epochs"

    @e2e_pytest_unit
    def test_get_unknown_task_epoch_name(self, mocker):
        mock_func1 = mocker.patch.object(TaskManager, "is_mmcv_framework_task")
        mock_func1.return_value = False
        mock_func2 = mocker.patch.object(TaskManager, "is_anomaly_framework_task")
        mock_func2.return_value = False

        task_manager = TaskManager(mocker.MagicMock())

        with pytest.raises(RuntimeError):
            task_manager.get_epoch_name()

    @e2e_pytest_unit
    @pytest.mark.parametrize("task", MMCV_TASK)
    def test_copy_weight(self, task: TaskType):
        task_manager = TaskManager(task)
        fake_model_weight = Path("temp_epoch_3.pth")
        with TemporaryDirectory() as src_dir, TemporaryDirectory() as det_dir:
            weight_in_src = src_dir / fake_model_weight
            weight_in_det = det_dir / fake_model_weight
            weight_in_src.write_text("fake")
            task_manager.copy_weight(src_dir, det_dir)

            assert weight_in_det.exists()

    @e2e_pytest_unit
    @pytest.mark.parametrize("task", MMCV_TASK)
    def test_get_latest_weight(self, task: TaskType):
        task_manager = TaskManager(task)

        with TemporaryDirectory() as work_dir:
            for i in range(1, 10):
                (work_dir / Path(f"epoch_{i}.pth")).write_text("fake")

            latest_model_weight = work_dir / Path("epoch_10.pth")
            latest_model_weight.write_text("fake")
            assert task_manager.get_latest_weight(work_dir) == str(latest_model_weight)


def get_template_path(template_dir: str) -> Path:
    task_config_dir = OTX_ROOT_PATH / "algorithms" / template_dir
    return list(task_config_dir.glob("**/template.yaml"))[0]


def make_task_env(template_path: str) -> TaskEnvironment:
    template = find_and_parse_model_template(template_path)
    return TaskEnvironment(template, None, create_conf_hp(template.hyper_parameters.data), MagicMock())


@pytest.fixture(scope="module")
def cls_template_path() -> str:
    return str(get_template_path("classification/configs"))


@pytest.fixture(scope="module")
def det_template_path() -> str:
    return str(get_template_path("detection/configs/detection"))


@pytest.fixture(scope="module")
def seg_template_path() -> str:
    return str(get_template_path("segmentation/configs"))


@pytest.fixture(scope="module")
def anomaly_template_path() -> str:
    return str(OTX_ROOT_PATH / "algorithms/anomaly/configs/classification/stfpm/template.yaml")


@pytest.fixture(scope="module")
def cls_task_env(cls_template_path):
    return make_task_env(cls_template_path)


@pytest.fixture(scope="module")
def det_task_env(det_template_path) -> TaskEnvironment:
    return make_task_env(det_template_path)


@pytest.fixture(scope="module")
def seg_task_env(seg_template_path) -> TaskEnvironment:
    return make_task_env(seg_template_path)


@pytest.fixture(scope="module")
def anomaly_task_env(anomaly_template_path) -> TaskEnvironment:
    return make_task_env(anomaly_template_path)


@pytest.fixture
def mmcv_task_env(cls_task_env, det_task_env, seg_task_env) -> List[TaskEnvironment]:
    return [cls_task_env, det_task_env, seg_task_env]


@pytest.fixture
def all_task_env(cls_task_env, det_task_env, seg_task_env, anomaly_task_env) -> List[TaskEnvironment]:
    return [cls_task_env, det_task_env, seg_task_env, anomaly_task_env]


@pytest.fixture
def mock_environment():
    MockTaskEnv = MagicMock(spec=TaskEnvironment)
    return MockTaskEnv()


@pytest.fixture(scope="module")
def action_template_path() -> str:
    return str(get_template_path("action"))


@pytest.fixture(scope="module")
def action_task_env(action_template_path) -> TaskEnvironment:
    return make_task_env(action_template_path)


class TestTaskEnvironmentManager:
    @pytest.fixture(autouse=True)
    def _make_mock_task_env(self, mock_environment):
        self._mock_environment = mock_environment

    @e2e_pytest_unit
    def test_init(self, all_task_env):
        for task_env in all_task_env:
            TaskEnvironmentManager(task_env)

    @e2e_pytest_unit
    def test_get_task(self, cls_task_env, det_task_env, seg_task_env):
        task_env = TaskEnvironmentManager(cls_task_env)
        assert task_env.get_task() == TaskType.CLASSIFICATION

        task_env = TaskEnvironmentManager(det_task_env)
        assert task_env.get_task() == TaskType.DETECTION

        task_env = TaskEnvironmentManager(seg_task_env)
        assert task_env.get_task() == TaskType.SEGMENTATION

    @e2e_pytest_unit
    def test_get_model_template(
        self, cls_task_env, det_task_env, seg_task_env, cls_template_path, det_template_path, seg_template_path
    ):
        task_env = TaskEnvironmentManager(cls_task_env)
        assert task_env.get_model_template() == find_and_parse_model_template(cls_template_path)

        task_env = TaskEnvironmentManager(det_task_env)
        assert task_env.get_model_template() == find_and_parse_model_template(det_template_path)

        task_env = TaskEnvironmentManager(seg_task_env)
        assert task_env.get_model_template() == find_and_parse_model_template(seg_template_path)

    @e2e_pytest_unit
    def test_get_model_template_path(
        self, cls_task_env, det_task_env, seg_task_env, cls_template_path, det_template_path, seg_template_path
    ):
        task_env = TaskEnvironmentManager(cls_task_env)
        assert task_env.get_model_template_path() == cls_template_path

        task_env = TaskEnvironmentManager(det_task_env)
        assert task_env.get_model_template_path() == det_template_path

        task_env = TaskEnvironmentManager(seg_task_env)
        assert task_env.get_model_template_path() == seg_template_path

    @e2e_pytest_unit
    def test_set_hyper_parameter_using_str_key(self):
        task_env = TaskEnvironmentManager(self._mock_environment)
        hyper_parameter = {"a.b.c.d": 1, "e.f.g.h": 2}

        task_env.set_hyper_parameter_using_str_key(hyper_parameter)

        env_hp = self._mock_environment.get_hyper_parameters()

        assert env_hp.a.b.c.d == hyper_parameter["a.b.c.d"]
        assert env_hp.e.f.g.h == hyper_parameter["e.f.g.h"]

    @e2e_pytest_unit
    def test_get_dict_type_hyper_parameter(self):
        learning_parameters = self._mock_environment.get_hyper_parameters().learning_parameters
        learning_parameters.parameters = ["a", "b"]
        learning_parameters.a = 1
        learning_parameters.b = 2

        task_env = TaskEnvironmentManager(self._mock_environment)
        dict_hp = task_env.get_dict_type_hyper_parameter()

        assert dict_hp["learning_parameters.a"] == 1
        assert dict_hp["learning_parameters.b"] == 2

    @e2e_pytest_unit
    @pytest.mark.parametrize("task", ALL_TASK)
    def test_get_max_epoch(self, task):
        max_epoch = 10
        self._mock_environment.model_template.task_type = task
        learning_parameters = self._mock_environment.get_hyper_parameters().learning_parameters
        setattr(learning_parameters, TaskManager(task).get_epoch_name(), max_epoch)

        task_env = TaskEnvironmentManager(self._mock_environment)

        assert task_env.get_max_epoch() == max_epoch

    @e2e_pytest_unit
    def test_save_mmcv_initial_weight(self, mmcv_task_env):
        for task_env in mmcv_task_env:
            task_env.model = None
            task_env = TaskEnvironmentManager(task_env)
            assert not task_env.save_initial_weight("fake_path")

    @e2e_pytest_unit
    def test_save_anomaly_initial_weight(self, mocker, anomaly_task_env):
        def mock_save_model_data(model, save_path: str):
            (Path(save_path) / "weights.pth").write_text("fake")

        mocker.patch.object(TaskEnvironmentManager, "get_train_task")
        mocker.patch("otx.cli.utils.hpo.save_model_data", mock_save_model_data)
        with TemporaryDirectory() as tmp_dir:
            anomaly_task_env.model = None
            task_env = TaskEnvironmentManager(anomaly_task_env)
            save_path = Path(tmp_dir) / "init.pth"
            assert task_env.save_initial_weight(str(save_path))
            assert save_path.exists()

    @e2e_pytest_unit
    def test_loaded_inital_weight(self, mocker, all_task_env):
        def mock_save_model_data(model, save_path: str):
            (Path(save_path) / "weights.pth").write_text("fake")

        mocker.patch.object(TaskEnvironmentManager, "get_train_task")
        mocker.patch("otx.cli.utils.hpo.save_model_data", mock_save_model_data)
        with TemporaryDirectory() as tmp_dir:
            for task_env in all_task_env:
                task_env.model = mocker.MagicMock()
                task_env = TaskEnvironmentManager(task_env)
                save_path = Path(tmp_dir) / "init.pth"
                assert task_env.save_initial_weight(str(save_path))
                assert save_path.exists()

    @e2e_pytest_unit
    def test_get_train_task(self, mocker, all_task_env):
        mock_func = mocker.patch("otx.cli.utils.hpo.get_impl_class")

        for task_env in all_task_env:
            mock_class = mocker.MagicMock()
            mock_func.return_vlaue = mock_class
            task_env = TaskEnvironmentManager(task_env)
            task_env.get_train_task()

            mock_class.assert_not_called()

    @e2e_pytest_unit
    def test_get_mmcv_batch_size_name(self, mmcv_task_env):
        for task_env in mmcv_task_env:
            task_env = TaskEnvironmentManager(task_env)
            assert task_env.get_batch_size_name() == "learning_parameters.batch_size"

    @e2e_pytest_unit
    def test_get_anomaly_batch_size_name(self, anomaly_task_env):
        task_env = TaskEnvironmentManager(anomaly_task_env)
        assert task_env.get_batch_size_name() == "learning_parameters.train_batch_size"

    @e2e_pytest_unit
    def test_load_model_weight(self, mocker, all_task_env):
        mock_func = mocker.patch("otx.cli.utils.hpo.read_model")

        for task_env in all_task_env:
            mock_class = mocker.MagicMock()
            mock_func.return_value = mock_class
            task_manager = TaskEnvironmentManager(task_env)
            task_manager.load_model_weight("fake", mocker.MagicMock())
            assert task_env.model == mock_class

    @e2e_pytest_unit
    def test_resume_model_weight(self, mocker, all_task_env):
        mock_func = mocker.patch("otx.cli.utils.hpo.read_model")

        for task_env in all_task_env:
            mock_class = mocker.MagicMock()
            mock_func.return_value = mock_class
            task_manager = TaskEnvironmentManager(task_env)
            task_manager.resume_model_weight("fake", mocker.MagicMock())
            assert task_env.model == mock_class
            assert mock_class.model_adapters["resume"]

    @e2e_pytest_unit
    def test_get_new_model_entity(self, all_task_env):
        for task_env in all_task_env:
            task_manager = TaskEnvironmentManager(task_env)
            model_entity = task_manager.get_new_model_entity()
            assert isinstance(model_entity, ModelEntity)

    @e2e_pytest_unit
    def test_set_epoch(self, all_task_env):
        epoch = 123
        for task_env in all_task_env:
            task_manager = TaskEnvironmentManager(task_env)
            task_manager.set_epoch(epoch)
            assert task_manager.get_max_epoch() == epoch


class TestHpoRunner:
    @e2e_pytest_unit
    def test_init(self, all_task_env):
        for task_env in all_task_env:
            HpoRunner(task_env, 100, 10, "fake_path")

    @e2e_pytest_unit
    @pytest.mark.parametrize("train_dataset_size,val_dataset_size", [(0, 10), (10, 0), (-1, -1)])
    def test_init_wrong_dataset_size(self, cls_task_env, train_dataset_size, val_dataset_size):
        with pytest.raises(ValueError):
            HpoRunner(cls_task_env, train_dataset_size, val_dataset_size, "fake_path", 4)

    @e2e_pytest_unit
    @pytest.mark.parametrize("hpo_time_ratio", [-3, 0])
    def test_init_wrong_hpo_time_ratio(self, cls_task_env, hpo_time_ratio):
        with pytest.raises(ValueError):
            HpoRunner(cls_task_env, 100, 10, "fake_path", hpo_time_ratio)

    @e2e_pytest_unit
    def test_run_hpo(self, mocker, cls_task_env):
        cls_task_env.model = None
        hpo_runner = HpoRunner(cls_task_env, 100, 10, "fake_path")
        mock_run_hpo_loop = mocker.patch("otx.cli.utils.hpo.run_hpo_loop")
        mock_hb = mocker.patch("otx.cli.utils.hpo.HyperBand")

        hpo_runner.run_hpo(mocker.MagicMock(), {"fake", "fake"})

        mock_run_hpo_loop.assert_called()  # call hpo_loop to run HPO
        mock_hb.assert_called()  # make hyperband

    @e2e_pytest_unit
    def test_run_hpo_w_dataset_smaller_than_batch(self, mocker, cls_task_env):
        cls_task_env.model = None
        hpo_runner = HpoRunner(cls_task_env, 2, 10, "fake_path")
        mock_run_hpo_loop = mocker.patch("otx.cli.utils.hpo.run_hpo_loop")
        mock_hb = mocker.patch("otx.cli.utils.hpo.HyperBand")

        hpo_runner.run_hpo(mocker.MagicMock(), {"fake", "fake"})

        mock_run_hpo_loop.assert_called()  # call hpo_loop to run HPO
        mock_hb.assert_called()  # make hyperband


class TestTrainer:
    @e2e_pytest_unit
    def test_init(self, mocker, cls_template_path):
        Trainer(
            hp_config={"configuration": {"iterations": 10}},
            report_func=mocker.stub(),
            model_template=find_and_parse_model_template(cls_template_path),
            data_roots={"fake": "fake"},
            task_type=TaskType.CLASSIFICATION,
            hpo_workdir="fake",
            initial_weight_name="fake",
            metric="fake",
        )

    @e2e_pytest_unit
    def test_run(self, mocker, cls_template_path):
        with TemporaryDirectory() as tmp_dir:
            # prepare
            trial_id = "1"
            weight_format = "epoch_{}.pth"
            hpo_workdir = Path(tmp_dir) / "hpo_dir"
            fake_project_path = Path(tmp_dir) / "fake_proejct"
            fake_project_path.mkdir(parents=True)
            for i in range(1, 5):
                (fake_project_path / weight_format.format(i)).write_text("fake")

            mock_get_train_task = mocker.patch.object(TaskEnvironmentManager, "get_train_task")
            mock_task = mocker.MagicMock()
            mock_task.project_path = str(fake_project_path)
            mock_get_train_task.return_value = mock_task

            mock_report_func = mocker.MagicMock()

            mocker.patch("otx.cli.utils.hpo.get_dataset_adapter")
            mocker.patch("otx.cli.utils.hpo.HpoDataset")

            # run
            trainer = Trainer(
                hp_config={"configuration": {"iterations": 10}, "id": trial_id},
                report_func=mock_report_func,
                model_template=find_and_parse_model_template(cls_template_path),
                data_roots=mocker.MagicMock(),
                task_type=TaskType.CLASSIFICATION,
                hpo_workdir=hpo_workdir,
                initial_weight_name="fake",
                metric="fake",
            )
            trainer.run()

            # check
            mock_report_func.assert_called_once_with(0, 0, done=True)  # finilize report
            assert hpo_workdir.exists()  # make a directory to copy weight
            for i in range(1, 5):  # check model weights are copied
                assert (hpo_workdir / "weight" / trial_id / weight_format.format(i)).exists()

        mock_task.train.assert_called()  # check task.train() is called


class TestHpoCallback:
    @e2e_pytest_unit
    def test_init(self, mocker):
        HpoCallback(mocker.MagicMock(), "fake", 3, mocker.MagicMock())

    @e2e_pytest_unit
    @pytest.mark.parametrize("max_epoch", [-3, 0])
    def test_init_wrong_max_epoch(self, mocker, max_epoch):
        with pytest.raises(ValueError):
            HpoCallback(mocker.MagicMock(), "fake", max_epoch, mocker.MagicMock())

    @e2e_pytest_unit
    def test_call(self, mocker):
        mock_report_func = mocker.MagicMock()

        hpo_call_back = HpoCallback(report_func=mock_report_func, metric="fake", max_epoch=50, task=mocker.MagicMock())
        hpo_call_back(progress=20, score=100)

        mock_report_func.assert_called_once_with(progress=10, score=100)

    @e2e_pytest_unit
    def test_call_and_get_stop_flag(self, mocker):
        mock_report_func = mocker.MagicMock()
        mock_report_func.return_value = TrialStatus.STOP
        mock_task = mocker.MagicMock()

        hpo_call_back = HpoCallback(report_func=mock_report_func, metric="fake", max_epoch=50, task=mock_task)
        hpo_call_back(progress=20, score=100)

        mock_task.cancel_training.assert_called_once_with()

    @e2e_pytest_unit
    def test_not_copy_report_func(self, mocker):
        mock_report_func = mocker.MagicMock()

        hpo_call_back = HpoCallback(report_func=mock_report_func, metric="fake", max_epoch=50, task=mocker.MagicMock())
        new_hpo_call_back = deepcopy(hpo_call_back)
        new_hpo_call_back(progress=20, score=100)

        mock_report_func.assert_called_once()


class TestHpoDataset:
    @e2e_pytest_unit
    def test_init(self, mocker):
        hpo_dataset = HpoDataset(fullset=mocker.MagicMock(), config={"train_environment": {"subset_ratio": 0.5}})
        assert hpo_dataset.subset_ratio == 0.5

    @e2e_pytest_unit
    @pytest.mark.parametrize("subset_ratio", [0.1, 0.5, 1])
    def test_get_subset(self, mocker, subset_ratio):
        mock_fullset = mocker.MagicMock()
        mock_fullset.get_subset.return_value = [i for i in range(10)]
        config = {"train_environment": {"subset_ratio": subset_ratio}}

        hpo_dataset = HpoDataset(fullset=mock_fullset, config=config)
        hpo_sub_dataset = hpo_dataset.get_subset(Subset.TRAINING)

        num_hpo_sub_dataset = len(hpo_sub_dataset)
        assert num_hpo_sub_dataset == round(10 * subset_ratio)

        for i in range(num_hpo_sub_dataset):
            hpo_sub_dataset[i]

    @e2e_pytest_unit
    def test_len_before_get_subset(self):
        hpo_dataset = HpoDataset(fullset=range(10), config={"train_environment": {"subset_ratio": 0.5}})
        assert len(hpo_dataset) == 10

    @e2e_pytest_unit
    def test_getitem_before_get_subset(self):
        hpo_dataset = HpoDataset(fullset=range(10), config={"train_environment": {"subset_ratio": 0.5}})

        for _ in hpo_dataset:
            pass


@e2e_pytest_unit
def test_run_hpo(mocker, mock_environment):
    with TemporaryDirectory() as tmp_dir:
        # prepare
        output = Path(tmp_dir) / "fake"
        mock_get_best_hpo_weight = mocker.patch("otx.cli.utils.hpo.get_best_hpo_weight")
        mock_get_best_hpo_weight.return_value = "mock_best_weight_path"

        def mock_run_hpo(*args, **kwargs):
            return {"config": {"a.b": 1, "c.d": 2}, "id": "1"}

        mock_hpo_runner_instance = mocker.MagicMock()
        mock_hpo_runner_instance.run_hpo.side_effect = mock_run_hpo
        mock_hpo_runner_class = mocker.patch("otx.cli.utils.hpo.HpoRunner")
        mock_hpo_runner_class.return_value = mock_hpo_runner_instance

        def mock_read_model(args1, path, arg2):
            return path

        mocker.patch("otx.cli.utils.hpo.read_model", mock_read_model)

        hpo_time_ratio = "4"
        mock_environment.model_template.task_type = TaskType.CLASSIFICATION

        # run
        environment = run_hpo(hpo_time_ratio, output, mock_environment, mocker.MagicMock(), mocker.MagicMock())

        # check
        mock_hpo_runner_instance.run_hpo.assert_called()  # Check that HpoRunner.run_hpo is called
        env_hp = environment.get_hyper_parameters()  # Check that best HP is applied well.
        assert env_hp.a.b == 1
        assert env_hp.c.d == 2
        assert environment.model == "mock_best_weight_path"  # check that best model weight is used


@e2e_pytest_unit
def test_run_hpo_not_supported_task(mocker, action_task_env):
    mock_hpo_runner_instance = mocker.MagicMock()
    mock_hpo_runner_class = mocker.patch("otx.cli.utils.hpo.HpoRunner")
    mock_hpo_runner_class.return_value = mock_hpo_runner_instance
    hpo_time_ratio = "4"
    output = "fake"

    run_hpo(hpo_time_ratio, output, action_task_env, mocker.MagicMock(), mocker.MagicMock())
    mock_hpo_runner_instance.run_hpo.assert_not_called()


@e2e_pytest_unit
def test_run_hpo_with_torchrun(mocker, mock_environment):
    # prepare
    output = "fake"
    mock_hpo_runner_instance = mocker.MagicMock()
    mock_hpo_runner_class = mocker.patch("otx.cli.utils.hpo.HpoRunner")
    mock_hpo_runner_class.return_value = mock_hpo_runner_instance
    hpo_time_ratio = "4"
    mock_environment.model_template.task_type = TaskType.CLASSIFICATION
    mock_os = mocker.patch.object(hpo, "os", return_value={"TORCHELASTIC_RUN_ID": "1234"})
    mock_os.environ = {"TORCHELASTIC_RUN_ID": "1234"}

    # run
    run_hpo(hpo_time_ratio, output, mock_environment, mocker.MagicMock(), mocker.MagicMock())
    mock_hpo_runner_instance.run_hpo.assert_not_called()


@e2e_pytest_unit
def test_get_best_hpo_weight():
    with TemporaryDirectory() as tmp_dir:
        # prepare
        hpo_dir = Path(tmp_dir) / "hpo"
        weight_path = hpo_dir / "weight"
        weight_path.mkdir(parents=True)

        score = {"score": {str(i): i for i in range(1, 11)}}
        bracket_0_dir = hpo_dir / "0"
        bracket_0_dir.mkdir(parents=True)
        for trial_num in range(2):
            with (bracket_0_dir / f"{trial_num}.json").open("w") as f:
                json.dump(score, f)
            trial_weight_path = weight_path / str(trial_num)
            trial_weight_path.mkdir(parents=True)
            for i in range(1, 11):
                (trial_weight_path / f"epoch_{i}.pth").write_text("fake")

        assert get_best_hpo_weight(hpo_dir, "1") == str(weight_path / "1" / "epoch_10.pth")


@e2e_pytest_unit
def test_get_best_hpo_weight_not_exist():
    with TemporaryDirectory() as tmp_dir:
        # prepare
        hpo_dir = Path(tmp_dir) / "hpo"
        weight_path = hpo_dir / "weight"
        weight_path.mkdir(parents=True)

        score = {"score": {str(i): i for i in range(1, 11)}}
        bracket_0_dir = hpo_dir / "0"
        bracket_0_dir.mkdir(parents=True)
        for trial_num in range(1):
            with (bracket_0_dir / f"{trial_num}.json").open("w") as f:
                json.dump(score, f)
            trial_weight_path = weight_path / str(trial_num)
            trial_weight_path.mkdir(parents=True)
            for i in range(1, 11):
                (trial_weight_path / f"epoch_{i}.pth").write_text("fake")

        assert get_best_hpo_weight(hpo_dir, "1") is None


@e2e_pytest_unit
def test_run_trial(mocker):
    mock_run = mocker.patch.object(Trainer, "run")
    run_trial(
        mocker.MagicMock(),
        mocker.MagicMock(),
        mocker.MagicMock(),
        mocker.MagicMock(),
        mocker.MagicMock(),
        mocker.MagicMock(),
        mocker.MagicMock(),
        mocker.MagicMock(),
    )

    mock_run.assert_called()



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-Lggggcense-Identifier: Apache-2.0
#

import pytest
import torch

from otx.core.ov.ops.type_conversions import ConvertV0
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestConvertV0:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 3, 256, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_convert_ov_type(self):
        with pytest.raises(NotImplementedError):
            ConvertV0.convert_ov_type("error")

        assert torch.uint8 == ConvertV0.convert_ov_type("u1")
        assert torch.uint8 == ConvertV0.convert_ov_type("u4")
        assert torch.uint8 == ConvertV0.convert_ov_type("u8")
        assert torch.int32 == ConvertV0.convert_ov_type("u32")
        assert torch.int64 == ConvertV0.convert_ov_type("u64")
        assert torch.int8 == ConvertV0.convert_ov_type("i4")
        assert torch.int8 == ConvertV0.convert_ov_type("i8")
        assert torch.int16 == ConvertV0.convert_ov_type("i16")
        assert torch.int32 == ConvertV0.convert_ov_type("i32")
        assert torch.int64 == ConvertV0.convert_ov_type("i64")
        assert torch.float16 == ConvertV0.convert_ov_type("f16")
        assert torch.float32 == ConvertV0.convert_ov_type("f32")
        assert torch.bool == ConvertV0.convert_ov_type("boolean")

    @e2e_pytest_unit
    def test_convert_torch_type(self):
        with pytest.raises(NotImplementedError):
            ConvertV0.convert_torch_type("error")

        assert "u8" == ConvertV0.convert_torch_type(torch.uint8)
        assert "i8" == ConvertV0.convert_torch_type(torch.int8)
        assert "i16" == ConvertV0.convert_torch_type(torch.int16)
        assert "i32" == ConvertV0.convert_torch_type(torch.int32)
        assert "i64" == ConvertV0.convert_torch_type(torch.int64)
        assert "f16" == ConvertV0.convert_torch_type(torch.float16)
        assert "f32" == ConvertV0.convert_torch_type(torch.float32)
        assert "boolean" == ConvertV0.convert_torch_type(torch.bool)

    @e2e_pytest_unit
    def test_forward(self):
        op = ConvertV0("dummy", shape=(1,), destination_type="f16")
        output = op(torch.randn(5, dtype=torch.float32))
        assert output.dtype == torch.float16



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import List

import pytest

from otx.api.entities.annotation import AnnotationSceneEntity, AnnotationSceneKind
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity, DatasetPurpose
from otx.api.entities.label import LabelEntity
from otx.api.entities.subset import Subset
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements
from tests.unit.api.entities.test_dataset_item import DatasetItemParameters


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestDatasetPurpose:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_purpose(self):
        """
        <b>Description:</b>
        Check DatasetPurpose Enum class elements

        <b>Expected results:</b>
        Test passes if DatasetPurpose Enum class length is equal to expected value, its elements have expected
        sequence numbers and value returned by __str__ method is equal to expected
        """
        assert len(DatasetPurpose) == 6
        assert DatasetPurpose.INFERENCE.value == 0
        assert str(DatasetPurpose.INFERENCE) == "INFERENCE"
        assert DatasetPurpose.TRAINING.value == 1
        assert str(DatasetPurpose.TRAINING) == "TRAINING"
        assert DatasetPurpose.EVALUATION.value == 2
        assert str(DatasetPurpose.EVALUATION) == "EVALUATION"
        assert DatasetPurpose.GENERATING_OUTPUT.value == 3
        assert str(DatasetPurpose.GENERATING_OUTPUT) == "GENERATING_OUTPUT"
        assert DatasetPurpose.TEMPORARY_DATASET.value == 4
        assert str(DatasetPurpose.TEMPORARY_DATASET) == "TEMPORARY_DATASET"
        assert DatasetPurpose.TASK_INFERENCE.value == 5
        assert str(DatasetPurpose.TASK_INFERENCE) == "TASK_INFERENCE"


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestDatasetEntity:
    @staticmethod
    def generate_random_image():
        return DatasetItemParameters.generate_random_image()

    @staticmethod
    def labels() -> List[LabelEntity]:
        return DatasetItemParameters.labels()

    @staticmethod
    def annotations_entity() -> AnnotationSceneEntity:
        return DatasetItemParameters().annotations_entity()

    @staticmethod
    def metadata():
        return DatasetItemParameters.metadata()

    @staticmethod
    def default_values_dataset_item() -> DatasetItemEntity:
        return DatasetItemParameters().default_values_dataset_item()

    @staticmethod
    def dataset_item() -> DatasetItemEntity:
        return DatasetItemParameters().dataset_item()

    def dataset(self) -> DatasetEntity:
        other_dataset_item = DatasetItemEntity(
            media=self.generate_random_image(),
            annotation_scene=self.annotations_entity(),
            metadata=self.metadata(),
            subset=Subset.VALIDATION,
        )
        items = [
            self.default_values_dataset_item(),
            self.dataset_item(),
            other_dataset_item,
        ]
        return DatasetEntity(items, DatasetPurpose.TEMPORARY_DATASET)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_entity_initialization(self):
        """
        <b>Description:</b>
        Check DatasetEntity class object initialization

        <b>Input data:</b>
        DatasetEntity class objects with specified "items" and "purpose" parameters

        <b>Expected results:</b>
        Test passes if "items" and "purpose" attributes of DatasetEntity object are equal to expected values

        <b>Steps</b>
        1. Check attributes of DatasetItemEntity object initialized with default optional parameters
        2. Check attributes of DatasetItemEntity object initialized with specified optional parameters
        """
        # Checking attributes of DatasetItemEntity object initialized with default optional parameters
        default_parameters_dataset = DatasetEntity()
        assert default_parameters_dataset._items == []
        assert default_parameters_dataset.purpose == DatasetPurpose.INFERENCE
        # Checking attributes of DatasetItemEntity object initialized with specified optional parameters
        items = [
            self.default_values_dataset_item(),
            self.dataset_item(),
            self.dataset_item(),
        ]
        purpose = DatasetPurpose.TEMPORARY_DATASET
        optional_parameters_dataset = DatasetEntity(items, purpose)
        assert optional_parameters_dataset._items == items
        assert optional_parameters_dataset.purpose == purpose

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_entity_purpose_setter(self):
        """
        <b>Description:</b>
        Check DatasetEntity class "purpose" setter

        <b>Input data:</b>
        DatasetEntity class objects with specified "items" and "purpose" parameters

        <b>Expected results:</b>
        Test passes if assigned value of "purpose" property is equal to expected

        <b>Steps</b>
        1. Check value returned by "purpose" property after using @purpose.setter for DatasetEntity with
        default optional parameters
        2. Check value returned by "purpose" property after using @purpose.setter for DatasetEntity initialized with
        specified optional parameters
        """
        # Checking "purpose" property after using @purpose.setter for DatasetEntity with default optional parameters
        default_parameters_dataset = DatasetEntity()
        expected_purpose = DatasetPurpose.TRAINING
        default_parameters_dataset.purpose = expected_purpose
        assert default_parameters_dataset.purpose == expected_purpose
        # Checking "purpose" property after using @purpose.setter for DatasetEntity with specified optional parameters
        optional_parameters_dataset = self.dataset()
        expected_purpose = DatasetPurpose.TASK_INFERENCE
        optional_parameters_dataset.purpose = expected_purpose
        assert optional_parameters_dataset.purpose == expected_purpose

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_entity_fetch(self):
        """
        <b>Description:</b>
        Check DatasetEntity class "_fetch" method

        <b>Input data:</b>
        DatasetEntity class object with specified "items" and "purpose" parameters

        <b>Expected results:</b>
        Test passes if value returned by "_fetch" method is equal to expected

        <b>Steps</b>
        1. Check value returned by "_fetch" method when list of DatasetItems indexes is specified as "key" parameter
        2. Check value returned by "_fetch" method when slice of DatasetItems indexes is specified as "key" parameter
        3. Check value returned by "_fetch" method when DatasetItem index is specified as "key" parameter
        4. Check TypeError exception is raised when unexpected type object is specified as "key" parameter
        """
        dataset = self.dataset()
        dataset_items = dataset._items
        # Checking  "_fetch" method when list of DatasetItems indexes is specified as "key" parameter
        assert dataset._fetch([0, 2]) == [dataset_items[0], dataset_items[2]]
        # Checking  "_fetch" method when slice of DatasetItems indexes is specified as "key" parameter
        assert dataset._fetch(slice(0, 3, 2)) == [dataset_items[0], dataset_items[2]]
        assert dataset._fetch(slice(-1, -4, -1)) == [
            dataset_items[2],
            dataset_items[1],
            dataset_items[0],
        ]
        # Checking  "_fetch" method when DatasetItem index is specified as "key" parameter
        assert dataset._fetch(1) == dataset_items[1]
        # Checking that TypeError exception is raised when unexpected type object is specified as "key" parameter
        with pytest.raises(TypeError):
            dataset._fetch(str)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_entity_repr(self):
        """
        <b>Description:</b>
        Check DatasetEntity class "__repr__" method

        <b>Input data:</b>
        DatasetEntity class objects with specified "items" and "purpose" parameters

        <b>Expected results:</b>
        Test passes if value returned by "__repr__" method is equal to expected

        <b>Steps</b>
        1. Check value returned by "__repr__" method for DatasetEntity with default optional parameters
        2. Check value returned by "__repr__" method for DatasetEntity with specified optional parameters
        """
        # Checking value returned by "__repr__" method for DatasetEntity with default optional parameters
        default_parameters_dataset = DatasetEntity()
        assert repr(default_parameters_dataset) == "DatasetEntity(items=[], purpose=INFERENCE)"
        # Checking value returned by "__repr__" method for DatasetEntity with specified optional parameters
        optional_parameters_dataset = self.dataset()
        assert (
            repr(optional_parameters_dataset) == f"DatasetEntity(items={optional_parameters_dataset._items}, "
            f"purpose=TEMPORARY_DATASET)"
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_entity_str(self):
        """
        <b>Description:</b>
        Check DatasetEntity class "__str__" method

        <b>Input data:</b>
        DatasetEntity class objects with specified "items" and "purpose" parameters

        <b>Expected results:</b>
        Test passes if value returned by "__str__" method is equal to expected

        <b>Steps</b>
        1. Check value returned by "__str__" method for DatasetEntity with default optional parameters
        2. Check value returned by "__str__" method for DatasetEntity with specified optional parameters
        """
        # Checking value returned by "__str__" method for DatasetEntity with default optional parameters
        default_parameters_dataset = DatasetEntity()
        assert str(default_parameters_dataset) == "DatasetEntity(size=0, purpose=INFERENCE)"
        # Checking value returned by "__str__" method for DatasetEntity with specified optional parameters
        optional_parameters_dataset = self.dataset()
        assert str(optional_parameters_dataset) == "DatasetEntity(size=3, purpose=TEMPORARY_DATASET)"

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_entity_len(self):
        """
        <b>Description:</b>
        Check DatasetEntity class "__len__" method

        <b>Input data:</b>
        DatasetEntity class objects with specified "items" and "purpose" parameters

        <b>Expected results:</b>
        Test passes if value returned by "__len__" method is equal to expected

        <b>Steps</b>
        1. Check value returned by "__len__" method for DatasetEntity with default optional parameters
        2. Check value returned by "__len__" method for DatasetEntity with specified optional parameters
        """
        # Checking value returned by "__str__" method for DatasetEntity with default optional parameters
        default_parameters_dataset = DatasetEntity()
        assert len(default_parameters_dataset) == 0
        # Checking value returned by "__str__" method for DatasetEntity with specified optional parameters
        optional_parameters_dataset = self.dataset()
        assert len(optional_parameters_dataset) == 3

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_entity_eq(self):
        """
        <b>Description:</b>
        Check DatasetEntity class "__eq__" method

        <b>Input data:</b>
        DatasetEntity class objects with specified "items" and "purpose" parameters

        <b>Expected results:</b>
        Test passes if value returned by "__eq__" method is equal to expected

        <b>Steps</b>
        1. Check value returned by "__eq__" method for equal DatasetEntity objects
        2. Check value returned by "__eq__" method for DatasetEntity objects with unequal length
        3. Check value returned by "__eq__" method for DatasetEntity objects with equal length, but unequal
        DatasetItem objects
        4. Check value returned by "__eq__" method for DatasetEntity objects with unequal "purpose" attributes
        5. Check value returned by "__eq__" method for comparing DatasetEntity object with object of different type
        """
        # Checking value returned by "__eq__" method for equal DatasetEntity objects
        items = [
            self.default_values_dataset_item(),
            self.dataset_item(),
            self.dataset_item(),
        ]
        purpose = DatasetPurpose.TEMPORARY_DATASET
        dataset = DatasetEntity(items, purpose)
        equal_dataset = DatasetEntity(items, purpose)
        assert dataset == equal_dataset
        # Checking value returned by "__eq__" method for DatasetEntity objects with unequal length
        unequal_items = list(items)
        unequal_items.pop(-1)
        unequal_dataset = DatasetEntity(unequal_items, purpose)
        assert dataset != unequal_dataset
        # Checking value returned by "__eq__" method for DatasetEntity objects with equal length, but unequal
        # DatasetItem objects
        unequal_items.append(self.dataset_item())
        unequal_dataset = DatasetEntity(unequal_items, purpose)
        assert dataset != unequal_dataset
        # Checking value returned by "__eq__" method for DatasetEntity objects with unequal "purpose" attributes
        unequal_dataset = DatasetEntity(items, DatasetPurpose.EVALUATION)
        assert dataset != unequal_dataset
        # Checking value returned by "__eq__" method for comparing DatasetEntity object with object of different type
        assert dataset != str

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_entity_add(self):
        """
        <b>Description:</b>
        Check DatasetEntity class "__add__" method

        <b>Input data:</b>
        DatasetEntity class object with specified "items" and "purpose" parameters

        <b>Expected results:</b>
        Test passes if DatasetEntity object returned by "__add__"" method is equal to expected

        <b>Steps</b>
        1. Check DatasetEntity object returned by "__add__"" method with DatasetEntity specified as "other" parameter
        2. Check DatasetEntity object returned by "__add__"" method with list of DatasetItemEntity objects specified
        as "other" parameter
        3. Check ValueError exception is raised when unexpected type object is specified in "other" parameter of
        "__add__" method
        """
        dataset = self.dataset()
        dataset_items = list(dataset._items)
        # Checking DatasetEntity object returned by "__add__"" method with DatasetEntity specified as "other" parameter
        other_dataset_items = [self.dataset_item(), self.dataset_item()]
        other_dataset = DatasetEntity(other_dataset_items, DatasetPurpose.TRAINING)
        new_dataset = dataset.__add__(other_dataset)
        assert new_dataset._items == dataset_items + other_dataset_items
        assert new_dataset.purpose == DatasetPurpose.TEMPORARY_DATASET
        # Checking DatasetEntity object returned by "__add__"" method with list of DatasetItemEntity objects specified
        # as "other" parameter
        items_to_add = [
            self.dataset_item(),
            self.dataset_item(),
            "unexpected type object",
        ]
        new_dataset = dataset.__add__(items_to_add)
        # Expected that str object will not be added to new_dataset._items
        assert new_dataset._items == dataset_items + items_to_add[0:2]
        assert new_dataset.purpose == DatasetPurpose.TEMPORARY_DATASET
        # Checking ValueError exception is raised when unexpected type object is specified in "other" parameter of
        # "__add__" method
        with pytest.raises(ValueError):
            dataset.__add__(str)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_entity_getitem(self):
        """
        <b>Description:</b>
        Check DatasetEntity class "__getitem__" method

        <b>Input data:</b>
        DatasetEntity class object with specified "items" and "purpose" parameters

        <b>Expected results:</b>
        Test passes if value returned by "__getitem__" method is equal to expected

        <b>Steps</b>
        1. Check value returned by "__getitem__" method when index is specified as "key" parameter
        2. Check value returned by "__getitem__" method when slice is specified as "key" parameter
        """
        dataset = self.dataset()
        dataset_items = dataset._items
        # Checking value returned by "__getitem__" method when index is specified as "key" parameter
        assert dataset[1] == dataset_items[1]
        # Checking value returned by "__getitem__" method when slice is specified as "key" parameter
        assert dataset[slice(0, 3, 2)] == [dataset_items[0], dataset_items[2]]
        assert dataset[slice(-1, -4, -1)] == [
            dataset_items[2],
            dataset_items[1],
            dataset_items[0],
        ]

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_entity_iter(self):
        """
        <b>Description:</b>
        Check DatasetEntity class "__iter__" method

        <b>Input data:</b>
        DatasetEntity class object with specified "items" and "purpose" parameters

        <b>Expected results:</b>
        Test passes if DatasetItemEntity returned by "__iter__" method is equal to expected
        """
        dataset = self.dataset()
        dataset_items = list(dataset._items)
        dataset_iterator = dataset.__iter__()
        expected_index = 0
        for expected_dataset_item in dataset_items:
            assert dataset_iterator.index == expected_index
            assert next(dataset_iterator) == expected_dataset_item
            expected_index += 1
        with pytest.raises(StopIteration):
            next(dataset_iterator)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_with_empty_annotations(self):
        """
        <b>Description:</b>
        Check DatasetEntity class "with_empty_annotations" method

        <b>Input data:</b>
        DatasetEntity class object with specified "items" and "purpose" parameters

        <b>Expected results:</b>
        Test passes if DatasetEntity object returned by "with_empty_annotations" method is equal to expected

        <b>Steps</b>
        1. Check DatasetEntity object returned by "with_empty_annotations" with non-specified "annotation_kind"
        parameter
        2. Check DatasetEntity object returned by "with_empty_annotations" with specified "annotation_kind" parameter
        """

        def check_empty_annotations_dataset(actual_dataset, expected_dataset, expected_kind):
            expected_items = expected_dataset._items
            actual_items = actual_dataset._items
            assert actual_dataset.purpose is expected_dataset.purpose
            for i in range(len(expected_items) - 1):
                actual_item = actual_items[i]
                expected_item = expected_items[i]
                assert actual_item.media is expected_item.media
                assert actual_item.annotation_scene.annotations == []
                assert actual_item.annotation_scene.kind == expected_kind
                assert actual_item.roi.id_ != expected_item.roi.id_
                assert actual_item.roi.shape is expected_item.roi.shape
                assert actual_item.roi.get_labels() == []
                assert actual_item.subset is expected_item.subset

        dataset = self.dataset()
        # Checking DatasetEntity object returned by "with_empty_annotations" with non-specified "annotation_kind"
        # parameter
        empty_annotations_dataset = dataset.with_empty_annotations()
        check_empty_annotations_dataset(empty_annotations_dataset, dataset, AnnotationSceneKind.PREDICTION)
        # Checking DatasetEntity object returned by "with_empty_annotations" with specified "annotation_kind" parameter
        kind = AnnotationSceneKind.ANNOTATION
        empty_annotations_dataset = dataset.with_empty_annotations(kind)
        check_empty_annotations_dataset(empty_annotations_dataset, dataset, kind)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_entity_get_subset(self):
        """
        <b>Description:</b>
        Check DatasetEntity class "get_subset" method

        <b>Input data:</b>
        DatasetEntity class object with specified "items" and "purpose" parameters

        <b>Expected results:</b>
        Test passes if DatasetEntity returned by "get_subset" method is equal to expected

        <b>Steps</b>
        1. Check DatasetEntity object returned by "get_subset" method with "subset" parameter that in items of base
        dataset
        2. Check DatasetEntity object returned by "get_subset" method with "subset" parameter that not in items of base
        dataset
        """
        validation_item = self.dataset_item()
        validation_item.subset = Subset.VALIDATION
        dataset = self.dataset()
        dataset._items.append(validation_item)
        # Checking DatasetEntity object returned by "get_subset" method with "subset" parameter that in items of base
        # dataset
        validation_dataset = dataset.get_subset(Subset.VALIDATION)
        assert validation_dataset.purpose is dataset.purpose
        assert validation_dataset._items == [dataset._items[2]] + [validation_item]
        # Checking DatasetEntity object returned by "get_subset" method with "subset" parameter that not in items of
        # base dataset
        empty_items_dataset = dataset.get_subset(Subset.UNLABELED)
        assert empty_items_dataset.purpose is dataset.purpose
        assert empty_items_dataset._items == []

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_entity_remove(self):
        """
        <b>Description:</b>
        Check DatasetEntity class "remove" method

        <b>Input data:</b>
        DatasetEntity class object with specified "items" and "purpose" parameters

        <b>Expected results:</b>
        Test passes if "items" attribute of DatasetEntity object is equal to expected after using "remove" method
        """
        dataset = self.dataset()
        dataset_items = list(dataset._items)
        # Removing DatasetItemEntity included in DatasetEntity
        dataset.remove(dataset_items[1])
        dataset_items.pop(1)
        assert dataset._items == dataset_items
        non_included_dataset_item = self.dataset_item()
        # Check that ValueError exception is raised when removing non-included DatasetItemEntity
        with pytest.raises(ValueError):
            dataset.remove(non_included_dataset_item)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_entity_append(self):
        """
        <b>Description:</b>
        Check DatasetEntity class "append" method

        <b>Input data:</b>
        DatasetEntity class object with specified "items" and "purpose" parameters

        <b>Expected results:</b>
        Test passes if "items" attribute of DatasetEntity object is equal to expected after using "append" method

        <b>Steps</b>
        1. Check "items" attribute of DatasetEntity object after adding new DatasetEntity object
        2. Check "items" attribute of DatasetEntity object after adding existing DatasetEntity object
        3. Check that ValueError exception is raised when appending DatasetEntity with "media" attribute is equal to
        "None"
        """
        dataset = self.dataset()
        expected_items = list(dataset._items)
        # Checking "items" attribute of DatasetEntity object after adding new DatasetEntity object
        item_to_add = self.dataset_item()
        dataset.append(item_to_add)
        expected_items.append(item_to_add)
        assert dataset._items == expected_items
        # Checking "items" attribute of DatasetEntity object after adding existing DatasetEntity object
        dataset.append(item_to_add)
        expected_items.append(item_to_add)
        assert dataset._items == expected_items
        # Checking that ValueError exception is raised when appending DatasetEntity with "media" is "None" attribute
        no_media_item = DatasetItemEntity(None, self.annotations_entity())
        with pytest.raises(ValueError):
            dataset.append(no_media_item)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_sort_items(self):
        """
        <b>Description:</b>
        Check DatasetEntity class "sort_items" method

        <b>Input data:</b>
        DatasetEntity class object with specified "items" and "purpose" parameters

        <b>Expected results:</b>
        Test passes if "items" attribute of DatasetEntity object is remained the same after using "sort_items" method
        """
        dataset = self.dataset()
        expected_items = list(dataset._items)
        dataset.sort_items()
        assert dataset._items == expected_items
        assert dataset.purpose == DatasetPurpose.TEMPORARY_DATASET

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_entity_remove_at_indices(self):
        """
        <b>Description:</b>
        Check DatasetEntity class "remove_at_indices" method

        <b>Input data:</b>
        DatasetEntity class object with specified "items" and "purpose" parameters

        <b>Expected results:</b>
        Test passes if "items" attribute of DatasetEntity object is equal to expected after using "remove_at_indices"
        method
        """
        dataset = self.dataset()
        expected_items = list(dataset._items)
        # Removing DatasetItemEntity included in DatasetEntity
        dataset.remove_at_indices([0, 2])
        expected_items.pop(2)
        expected_items.pop(0)
        assert dataset._items == expected_items
        # Check that IndexError exception is raised when removing DatasetItemEntity with non-included index
        with pytest.raises(IndexError):
            dataset.remove_at_indices([20])

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dataset_entity_get_labels(self):
        """
        <b>Description:</b>
        Check DatasetEntity class "get_labels" method

        <b>Input data:</b>
        DatasetEntity class object with specified "items" and "purpose" parameters

        <b>Expected results:</b>
        Test passes if list returned by "get_labels" method is equal to expected

         <b>Steps</b>
        1. Check list returned by "get_labels" method with "include_empty" parameter is "False"
        2. Check list returned by "get_labels" method with "include_empty" parameter is "True"
        """
        labels = self.labels()
        detection_label = labels[0]
        segmentation_empty_label = labels[1]
        dataset = self.dataset()
        # Checking list returned by "get_labels" method with "include_empty" parameter is "False"
        assert dataset.get_labels() == [detection_label]
        # Checking list returned by "get_labels" method with "include_empty" parameter is "True"
        actual_empty_labels = dataset.get_labels(include_empty=True)
        assert len(actual_empty_labels) == 2
        assert isinstance(actual_empty_labels, list)
        assert segmentation_empty_label in actual_empty_labels
        assert detection_label in actual_empty_labels



"""Generation-related module for otx.core.ov.ops."""
# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass

import torch

from otx.core.ov.ops.builder import OPS
from otx.core.ov.ops.op import Attribute, Operation
from otx.core.ov.ops.type_conversions import _ov_to_torch


@dataclass
class RangeV4Attribute(Attribute):
    """RangeV4Attribute class."""

    output_type: str


@OPS.register()
class RangeV4(Operation[RangeV4Attribute]):
    """RangeV4 class."""

    TYPE = "Range"
    VERSION = 4
    ATTRIBUTE_FACTORY = RangeV4Attribute

    def forward(self, start, stop, step):
        """RangeV4's forward function."""
        dtype = _ov_to_torch[self.attrs.output_type]
        return torch.arange(
            start=start,
            end=stop,
            step=step,
            dtype=dtype,
            device=start.device,
            requires_grad=False,
        )



"""Unit-Test Case for otx.cli.builder.builder."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
from pathlib import Path

import mmcv
import pytest
from mmcv.utils import Registry
from torch import nn

from otx.algorithms.common.adapters.mmcv.utils.config_utils import MPAConfig
from otx.cli.builder.builder import (
    Builder,
    get_backbone_out_channels,
    update_backbone_args,
    update_channels,
)
from otx.cli.utils.importing import get_otx_root_path
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestOTXCLIBuilder:
    """Check Builder's function is working well.

    1. Check "Builder.build_backbone_config" function that generate backbone configuration file is working well
    <Steps>
        1. Generate backbone config file (mmcls.MMOVBackbone)
        2. Raise ValueError with wrong output_path

    2. Check "Builder.merge_backbone" function that update model config with new backbone is working well
    <Steps>
        1. Update model config with mmcls.ResNet backbone (default model.backbone: otx.OTXEfficientNet)
        2. Raise ValueError with wrong model_config_path
        3. Raise ValueError with wrong backbone_config_path
        4. Update model config without backbone's out_indices
        5. Update model config with backbone's pretrained path
    """

    @pytest.fixture(autouse=True)
    def setup(self, tmp_dir_path: str) -> None:
        self.otx_builder = Builder()
        self.otx_root = get_otx_root_path()
        self.tmp_dir_path = tmp_dir_path if isinstance(tmp_dir_path, Path) else Path(tmp_dir_path)

    @e2e_pytest_unit
    @pytest.mark.parametrize("backbone_type", ["mmcls.MMOVBackbone"])
    def test_builder_build_backbone_config_generate_backbone(self, backbone_type: str) -> None:
        """Generate backbone config file (mmcls.MMOVBackbone)."""
        tmp_backbone_path = self.tmp_dir_path / "backbone.yaml"
        self.otx_builder.build_backbone_config(backbone_type, tmp_backbone_path)
        assert tmp_backbone_path.exists()
        backbone_config = mmcv.load(str(tmp_backbone_path))
        assert backbone_config["backbone"]["type"] == backbone_type

    @e2e_pytest_unit
    @pytest.mark.parametrize("backbone_type", ["mmcls.MMOVBackbone"])
    def test_builder_build_backbone_config_abnormal_output_path(self, backbone_type: str) -> None:
        """Raise ValueError with wrong output_path."""
        tmp_backbone_path = self.tmp_dir_path / "wrong.path"
        with pytest.raises(ValueError):
            self.otx_builder.build_backbone_config(backbone_type, tmp_backbone_path)

    @e2e_pytest_unit
    def test_builder_merge_backbone_abnormal_model_path(self) -> None:
        """Raise ValueError with wrong model_config_path."""
        workspace_path = self.tmp_dir_path / "test_builder_merge_backbone"
        tmp_backbone_path = workspace_path / "backbone.yaml"
        with pytest.raises(ValueError):
            self.otx_builder.merge_backbone("unexpected", tmp_backbone_path)

    @e2e_pytest_unit
    def test_builder_merge_backbone_abnormal_backbone_path(self) -> None:
        """Raise ValueError with wrong backbone_config_path."""
        workspace_path = self.tmp_dir_path / "test_builder_merge_backbone"
        tmp_model_path = workspace_path / "model.py"
        with pytest.raises(ValueError):
            self.otx_builder.merge_backbone(tmp_model_path, "unexpected")

    @e2e_pytest_unit
    def test_builder_merge_backbone(self, mocker) -> None:
        """Update model config without backbone's out_indices."""
        mocker.patch("otx.cli.builder.builder.Path.exists", return_value=True)
        mock_backbone_config = {
            "backbone": {"type": "torchvision.resnet18", "use_out_indices": True, "out_indices": [0, 1, 2]}
        }
        mock_mmcv_load = mocker.patch("otx.cli.builder.builder.mmcv.load", return_value=mock_backbone_config)
        mock_model_config = mocker.MagicMock()
        mock_model_config.model.backbone = {"out_indices"}
        mock_config_from_file = mocker.patch(
            "otx.cli.builder.builder.MPAConfig.fromfile", return_value=mock_model_config
        )
        mock_get_backbone_registry = mocker.patch(
            "otx.cli.builder.builder.get_backbone_registry", return_value=(None, ["otx"])
        )
        mock_backbone = mocker.MagicMock()
        mock_build_from_cfg = mocker.patch("otx.cli.builder.builder.build_from_cfg", return_value=mock_backbone)
        mock_get_backbone_out_channels = mocker.patch(
            "otx.cli.builder.builder.get_backbone_out_channels", return_value=[0, 1, 2]
        )
        mock_update_channels = mocker.patch("otx.cli.builder.builder.update_channels", return_value=None)

        Builder().merge_backbone("fake_model_path", "fake_backbone_path")

        mock_mmcv_load.assert_called_once()
        mock_config_from_file.assert_called_once()
        mock_get_backbone_registry.assert_called_once()
        mock_build_from_cfg.assert_called_once()
        mock_get_backbone_out_channels.assert_called_once()
        mock_update_channels.assert_called_once()


class MockBackbone(nn.Module):
    def __init__(self) -> None:
        super(MockBackbone, self).__init__()

    def forward(self, x):
        return x


class MockRegistry(Registry):
    def __init__(self, name: str, parent: Registry = None, scope: str = None) -> None:
        super(MockRegistry, self).__init__(name=name, parent=parent, scope=scope)


class TestOTXBuilderUtils:
    """Check util function in builder.py is working well.

    1. Check "get_backbone_out_channels" function that checking backbone's output channels is working well.
    <Steps>
        1. Check default input_size = 64
        2. Check backbone input_size change to 128

    2. Check "update_backbone_args" function is working.
    <Steps>
        1. Update backbone args (Case without Required Args)
        2. Update required Args in Backbone (Check Missing Args)
        3. Update Args with out_indices
        4. Update backbone using the backbone name from the backbone list (Check updating with options)
        5. Update backbone using the backbone name from the backbone list (Check updating without options)
        6. Raise ValueError with unexpected backbone

    3. Check "update_channels" function is working.
    <Steps>
        1. Remove model.neck.in_channels (GlobalAveragePooling)
        2. Update model.neck.in_channels
        3. Update model.decode_head.in_channels & in_index (segmentation case)
        4. Update model.head.in_channels
        5. Raise NotImplementedError with unexpected model key
    """

    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.registry = MockRegistry(name="test")
        self.backbone = MockBackbone

    @e2e_pytest_unit
    def test_get_backbone_out_channels(self) -> None:
        """Check "get_backbone_out_channels" function that checking backbone's output channels is working well.

        1. Check default input_size = 64
        2. Check backbone input_size change to 128
        """

        backbone = MockBackbone()
        expected_result = [64, 64]
        out_channels = get_backbone_out_channels(backbone)
        assert out_channels == expected_result

        backbone.input_size = 128
        expected_result = [128, 128]
        out_channels = get_backbone_out_channels(backbone)
        assert out_channels == expected_result

    @e2e_pytest_unit
    def test_update_backbone_args_without_required_args(self) -> None:
        """Update backbone args (Case without Required Args)."""

        def mock_init(self, a=1, b=2):
            super(MockBackbone, self).__init__()

        self.backbone.__init__ = mock_init
        self.registry.register_module(module=self.backbone)

        backbone_config = {"type": "MockBackbone"}
        inputs = {"backbone_config": backbone_config, "registry": self.registry, "backend": "mmseg"}
        results = update_backbone_args(**inputs)
        expected_results = []
        assert results == expected_results

    @e2e_pytest_unit
    def test_update_backbone_args_required_args(self) -> None:
        """Update required Args in Backbone (Check Missing Args)."""

        def mock_init(self, depth, a=1, b=2):
            super(MockBackbone, self).__init__()

        self.backbone.__init__ = mock_init
        self.registry.register_module(module=self.backbone, force=True)
        backbone_config = {"type": "MockBackbone"}
        inputs = {"backbone_config": backbone_config, "registry": self.registry, "backend": "mmseg"}
        results = update_backbone_args(**inputs)
        expected_results = ["depth"]
        assert results == expected_results

    @e2e_pytest_unit
    def test_update_backbone_args_with_out_indices(self) -> None:
        """Update Args with out_indices."""
        backbone_config = {"type": "MockBackbone", "out_indices": (0, 1, 2, 3)}
        self.registry.register_module(module=self.backbone, force=True)
        inputs = {"backbone_config": backbone_config, "registry": self.registry, "backend": "mmseg"}
        results = update_backbone_args(**inputs)
        expected_results = ["depth"]
        assert results == expected_results
        assert "use_out_indices" in backbone_config

    @e2e_pytest_unit
    def test_update_backbone_args_with_option(self) -> None:
        """Update backbone using the backbone name from the backbone list (Check updating with options)."""
        child_registry = MockRegistry(name="mmseg", parent=self.registry, scope="mmseg")
        backbone_config = {"type": "mmseg.ResNet"}
        child_registry.register_module(name="ResNet", module=self.backbone, force=True)
        inputs = {"backbone_config": backbone_config, "registry": self.registry, "backend": "mmseg"}
        results = update_backbone_args(**inputs)
        expected_results = []
        assert results == expected_results
        expected_depth = 18
        assert "depth" in backbone_config
        assert backbone_config["depth"] == expected_depth

    @e2e_pytest_unit
    def test_update_backbone_args_without_options(self) -> None:
        """Update backbone using the backbone name from the backbone list (Check updating without options)."""

        def mock_init(self, extra):
            super(MockBackbone, self).__init__()

        backbone_config = {"type": "mmseg.HRNet"}
        self.backbone.__init__ = mock_init
        child_registry = MockRegistry(name="mmseg", parent=self.registry, scope="mmseg")
        child_registry.register_module(name="HRNet", module=self.backbone, force=True)
        inputs = {"backbone_config": backbone_config, "registry": self.registry, "backend": "mmseg"}
        results = update_backbone_args(**inputs)
        expected_results = ["extra"]
        assert results == expected_results
        expected_extra_value = "!!!!!!!!!!!INPUT_HERE!!!!!!!!!!!"
        assert "extra" in backbone_config
        assert backbone_config["extra"] == expected_extra_value

    @e2e_pytest_unit
    def test_update_backbone_args_abnormal_backbone_type(self) -> None:
        """Raise ValueError with unexpected backbone."""
        backbone_config = {"type": "unexpected"}
        inputs = {"backbone_config": backbone_config, "registry": self.registry, "backend": "mmseg"}
        with pytest.raises(ValueError):
            update_backbone_args(**inputs)

    @e2e_pytest_unit
    def test_update_channels_neck(self) -> None:
        """Remove model.neck.in_channels."""
        cfg_dict = {"model": {"neck": {"type": "GlobalAveragePooling", "in_channels": 100}}}
        model_config = MPAConfig(cfg_dict=cfg_dict)
        update_channels(model_config, -1)
        assert "in_channels" not in model_config.model.neck

        cfg_dict = {"model": {"neck": {"type": "TestNeck", "in_channels": 100}}}
        model_config = MPAConfig(cfg_dict=cfg_dict)
        update_channels(model_config, -1)
        assert model_config.model.neck.in_channels == -1

    @e2e_pytest_unit
    def test_update_channels_decode_head(self) -> None:
        """Update model.decode_head.in_channels & in_index (segmentation case)."""
        cfg_dict = {"model": {"decode_head": {"in_channels": (0, 1, 2), "in_index": (0, 1, 2)}}}
        out_channels = (10, 20, 30, 40)
        model_config = MPAConfig(cfg_dict=cfg_dict)
        update_channels(model_config, out_channels)
        assert model_config.model.decode_head.in_index == list(range(len(out_channels)))
        assert model_config.model.decode_head.in_channels == out_channels

    @e2e_pytest_unit
    def test_update_channels_head(self) -> None:
        """Update model.head.in_channels."""
        out_channels = (10, 20, 30, 40)
        cfg_dict = {"model": {"head": {"in_channels": (0, 1, 2)}}}
        model_config = MPAConfig(cfg_dict=cfg_dict)
        update_channels(model_config, out_channels)
        assert model_config.model.head.in_channels == out_channels

    @e2e_pytest_unit
    def test_update_channels_abnormal_inputs(self) -> None:
        """Raise NotImplementedError with unexpected model key."""
        out_channels = (10, 20, 30, 40)
        cfg_dict = {"model": {"unexpected": {"in_channels": (0, 1, 2)}}}
        model_config = MPAConfig(cfg_dict=cfg_dict)
        with pytest.raises(NotImplementedError):
            update_channels(model_config, out_channels)
